# Copyright (c) 2010-2016 Bo Lin
# Copyright (c) 2010-2016 Yanhong Annie Liu
# Copyright (c) 2010-2016 Stony Brook University
# Copyright (c) 2010-2016 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import builtins

from ast import *
from collections import abc

from da import common
from da.compiler import dast
from da.compiler.utils import printe, printw, printd, Namespace

# DistAlgo keywords
KW_ENTRY_POINT = "main"
KW_PROCESS_DEF = "process"
KW_PROCESS_ENTRY_POINT = "run"
KW_CONFIG = "config"
KW_SETUP = "setup"
KW_START = "start"
KW_RECV_QUERY = "received"
KW_SENT_QUERY = "sent"
KW_RECV_EVENT = "receive"
KW_SENT_EVENT = "sent"
KW_MSG_PATTERN = "msg"
KW_EVENT_SOURCE = "from_"
KW_EVENT_DESTINATION = "to"
KW_EVENT_TIMESTAMP = "clk"
KW_EVENT_LABEL = "at"
KW_DECORATOR_LABEL = "labels"
KW_EXISTENTIAL_QUANT = "some"
KW_UNIVERSAL_QUANT = "each"
KW_AGGREGATE_SIZE = "len"
KW_AGGREGATE_MIN = "min"
KW_AGGREGATE_MAX = "max"
KW_AGGREGATE_SUM = "sum"
KW_COMP_SET = "setof"
KW_COMP_TUPLE = "tupleof"
KW_COMP_LIST = "listof"
KW_COMP_DICT = "dictof"
KW_AWAIT = "await"
KW_AWAIT_TIMEOUT = "timeout"
KW_SEND = "send"
KW_SEND_TO = KW_EVENT_DESTINATION
KW_PRINT = "output"
KW_LEVEL = "level"
KW_SEP= "sep"
KW_SELF = "self"
KW_TRUE = "True"
KW_FALSE = "False"
KW_NULL = "None"
KW_SUCH_THAT = "has"
KW_RESET = "reset"
KW_INC_VERB = "_INC_"

##########################
# Helper functions:
##########################

def is_setup_func(node):
    """Returns True if this node defines a function named 'setup'."""

    return (isinstance(node, FunctionDef) and
            node.name == KW_SETUP)

def extract_label(node):
    """Returns the label name specified in 'node', or None if 'node' is not a
    label.
    """
    if (isinstance(node, UnaryOp) and
            isinstance(node.op, USub) and
            isinstance(node.operand, UnaryOp) and
            isinstance(node.operand.op, USub) and
            isinstance(node.operand.operand, Name)):
        return node.operand.operand.id
    else:
        return None

def daast_from_file(filename, args=None):
    """Generates DistAlgo AST from source file.

    'filename' is the filename of source file. Optional argument 'args' is a
    Namespace object containing the command line parameters for the compiler.
    Returns the generated DistAlgo AST.

    """
    try:
        with open(filename, 'r') as infd:
            global InputSize
            src = infd.read()
            InputSize = len(src)
            return daast_from_str(src, filename, args)
    except Exception as e:
        print(type(e).__name__, ':', str(e), file=sys.stderr)
        raise e
    return None

def daast_from_str(src, filename='<str>', args=None):
    """Generates DistAlgo AST from source string.

    'src' is the DistAlgo source string to parse. Optional argument 'filename'
    specifies the filename that appears in error messages, defaults to
    '<str>'. Optional argument 'args' is a Namespace object containing the
    command line parameters for the compiler. Returns the generated DistAlgo
    AST.

    """
    try:
        dt = Parser(filename, args)
        rawast = parse(src, filename)
        dt.visit(rawast)
        sys.stderr.write("%s compiled with %d errors and %d warnings.\n" %
                     (filename, dt.errcnt, dt.warncnt))
        if dt.errcnt == 0:
            return dt.program
    except SyntaxError as e:
        sys.stderr.write("%s:%d:%d: SyntaxError: %s" % (e.filename, e.lineno,
                                                    e.offset, e.text))
    return None

##########
# Operator mappings:
##########
NegatedOperators = {
    NotEq : dast.EqOp,
    IsNot : dast.IsOp,
    NotIn : dast.InOp
}

OperatorMap = {
    Add : dast.AddOp,
    Sub : dast.SubOp,
    Mult : dast.MultOp,
    Div : dast.DivOp,
    Mod : dast.ModOp,
    Pow : dast.PowOp,
    LShift : dast.LShiftOp,
    RShift : dast.RShiftOp,
    BitOr : dast.BitOrOp,
    BitXor : dast.BitXorOp,
    BitAnd : dast.BitAndOp,
    FloorDiv : dast.FloorDivOp,

    Eq : dast.EqOp,
    NotEq: dast.NotEqOp,
    Lt : dast.LtOp,
    LtE : dast.LtEOp,
    Gt : dast.GtOp,
    GtE : dast.GtEOp,
    Is : dast.IsOp,
    IsNot : dast.IsNotOp,
    In : dast.InOp,
    NotIn : dast.NotInOp,

    USub : dast.USubOp,
    UAdd : dast.UAddOp,
    Invert : dast.InvertOp,

    And : dast.AndOp,
    Or : dast.OrOp
}
# New matrix multiplication operator since 3.5:
if sys.version_info > (3, 5):
    OperatorMap[MatMult] = dast.MatMultOp

# FIXME: is there a better way than hardcoding these?
KnownUpdateMethods = {
    "add", "append", "extend", "update",
    "insert", "reverse", "sort",
    "delete", "remove", "pop", "clear", "discard"
}

ValidResetTypes = {"Received", "Sent", ""}
ApiMethods = frozenset(common.api_registry.keys())
BuiltinMethods = frozenset(common.builtin_registry.keys())
InternalMethods = frozenset(common.internal_registry.keys())
PythonBuiltins = dir(builtins)
NodeProcName = "Node_"

AggregateMap = {
    KW_AGGREGATE_MAX  : dast.MaxExpr,
    KW_AGGREGATE_MIN  : dast.MinExpr,
    KW_AGGREGATE_SIZE : dast.SizeExpr,
    KW_AGGREGATE_SUM  : dast.SumExpr
}
ComprehensionTypes = {KW_COMP_SET, KW_COMP_TUPLE, KW_COMP_DICT, KW_COMP_LIST}
EventKeywords = {KW_EVENT_DESTINATION, KW_EVENT_SOURCE, KW_EVENT_LABEL,
                 KW_EVENT_TIMESTAMP}
Quantifiers = {KW_UNIVERSAL_QUANT, KW_EXISTENTIAL_QUANT}

##########
# Exceptions:
class MalformedStatementError(Exception):
    def __init__(self, reason=None, node=None, name=None, msg=None):
        super().__init__(reason, node, name, msg)
        self.reason = reason
        self.node = node
        self.name = name
        self.msg = msg

##########
# Name context types:
class NameContext:
    def __init__(self, node, type=None):
        self.node = node        # The node containing this context
        self.type = type        # Optional, for type analysis

class Assignment(NameContext): pass
class Update(NameContext): pass
class Read(NameContext): pass
class IterRead(Read): pass
class FunCall(NameContext): pass
class Delete(NameContext): pass
class AttributeLookup(NameContext): pass
class SubscriptLookup(NameContext): pass
class PatternContext(NameContext): pass
class Existential(NameContext): pass
class Universal(NameContext): pass

##########

class PatternParser(NodeVisitor):
    """Parses a pattern.
    """

    def __init__(self, parser, literal=False):
        self._parser = parser
        self.namescope = parser.current_scope
        self.parent_node = parser.current_parent
        self.current_query = parser.current_query
        self.use_object_style = parser.get_option('enable_object_pattern',
                                                  default=False)
        self.use_top_semantic = parser.get_option('use_top_semantic',
                                                  default=False)
        self.literal = literal

    def visit(self, node):
        if isinstance(node, Name):
            return self.visit_Name(node)
        elif isinstance(node, Tuple):
            return self.visit_Tuple(node)
        elif isinstance(node, List):
            return self.visit_List(node)

        # Parse general expressions:
        expr = self._parser.visit(node)
        if isinstance(expr, dast.ConstantExpr):
            return dast.ConstantPattern(self.parent_node, node, value=expr)
        else:
            return dast.BoundPattern(self.parent_node, node, value=expr)

    def is_bound(self, name):
        n = self.namescope.find_name(name)
        if n is not None and self.current_query is not None:
            if self.use_top_semantic:
                return n in self.current_query.top_level_query.boundvars
            else:
                return n in self.current_query.boundvars
        else:
            return False

    def is_free(self, name):
        n = self.namescope.find_name(name)
        if n is not None and self.current_query is not None:
            if self.use_top_semantic:
                return n in self.current_query.top_level_query.freevars
            else:
                return n in self.current_query.freevars
        else:
            return False

    def visit_Name(self, node):
        if self._parser.current_process is not None and \
           node.id == KW_SELF:
            return dast.ConstantPattern(
                self.parent_node, node,
                value=dast.SelfExpr(self.parent_node, node))
        elif node.id == KW_TRUE:
            return dast.ConstantPattern(
                self.parent_node, node,
                value=dast.TrueExpr(self.parent_node, node))
        elif node.id == KW_FALSE:
            return dast.ConstantPattern(
                self.parent_node, node,
                value=dast.FalseExpr(self.parent_node, node))
        elif node.id == KW_NULL:
            return dast.ConstantPattern(
                self.parent_node, node,
                value=dast.NoneExpr(self.parent_node, node))
        elif self.literal:
            name = node.id
            n = self.namescope.find_name(name)
            if n is None:
                n = self.namescope.add_name(name)
            pat = dast.BoundPattern(self.parent_node, node, value=n)
            n.add_read(pat)
            return pat

        name = node.id
        if name == "_":
            # Wild card
            return dast.FreePattern(self.parent_node, node)
        elif name.startswith("_"):
            # Bound variable:
            name = node.id[1:]
            n = self.namescope.find_name(name)
            if n is None:
                self._parser.warn(
                    ("new variable '%s' introduced by bound pattern." % name),
                    node)
                n = self.namescope.add_name(name)
            pat = dast.BoundPattern(self.parent_node, node, value=n)
            n.add_read(pat)
            return pat
        else:
            # Could be free or bound:
            name = node.id
            if self.is_bound(name):
                self._parser.debug("[PatternParser] reusing bound name " +
                                   name, node)
                n = self.namescope.find_name(name)
                pat = dast.BoundPattern(self.parent_node, node, value=n)
                n.add_read(pat)
            else:
                self._parser.debug("[PatternParser] free name " + name, node)
                n = self.namescope.find_name(name)
                if n is None:
                    # New name:
                    n = self.namescope.add_name(name)
                    pat = dast.FreePattern(self.parent_node, node, value=n)
                    n.add_assignment(self.parent_node)
                else:
                    pat = dast.FreePattern(self.parent_node, node, value=n)
                    if self.is_free(name):
                        # Existing free var:
                        n.add_read(pat)
                    else:
                        n.add_assignment(self.parent_node)
            return pat

    def visit_Str(self, node):
        return dast.ConstantPattern(
            self.parent_node, node,
            value=dast.ConstantExpr(self.parent_node, node, node.s))

    def visit_Bytes(self, node):
        return dast.ConstantPattern(
            self.parent_node, node,
            value=dast.ConstantExpr(self.parent_node, node, node.s))

    def visit_Num(self, node):
        return dast.ConstantPattern(
            self.parent_node, node,
            value=dast.ConstantExpr(self.parent_node, node, node.n))

    def visit_Tuple(self, node):
        return dast.TuplePattern(
            self.parent_node, node,
            value=[self.visit(e) for e in node.elts])

    def visit_List(self, node):
        return dast.ListPattern(
            self.parent_node, node,
            value=[self.visit(e) for e in node.elts])

    def visit_Call(self, node):
        if not self.use_object_style:
            return self.generic_visit(node)

        if not isinstance(node.func, Name): return None
        elts = [dast.ConstantPattern(
            self.parent_node, node,
            value=dast.ConstantExpr(self.parent_node,
                                    node.func,
                                    value=node.func.id))]
        for e in node.args:
            elts.append(self.visit(e))
        return dast.TuplePattern(self.parent_node, node,
                                 value=elts)

class Pattern2Constant(NodeVisitor):
    def __init__(self, parent):
        super().__init__()
        self.stack = [parent]

    @property
    def current_parent(self):
        return self.stack[-1]

    def visit_ConstantPattern(self, node):
        expr = node.value.clone()
        expr._parent = self.current_parent
        return expr

    def visit_BoundPattern(self, node):
        return das.SimpleExpr(node.parent, value=node.value)

    def visit_FreePattern(self, node):
        # This really shouldn't happen
        mesg = "Can not convert FreePattern to constant!"
        printe(mesg, node.lineno, node.col_offset)
        return None

    def visit_TuplePattern(self, node):
        expr = dast.TupleExpr(self.current_parent)
        self.stack.append(expr)
        expr.subexprs = [self.visit(e) for e in node.value]
        self.stack.pop()
        if None in expr.subexprs:
            expr = None
        return expr

    def visit_ListPattern(self, node):
        expr = dast.ListExpr(self.current_parent)
        self.stack.append(expr)
        expr.subexprs = [self.visit(e) for e in node.value]
        self.stack.pop()
        if None in expr.subexprs:
            expr = None
        return expr

class PatternFinder(NodeVisitor):
    def __init__(self):
        self.found = False

    # It's a pattern if it has bound variables:
    def visit_Name(self, node):
        if node.id.startswith("_"):
            self.found = True

    # It's also a pattern if it contains constants:
    def visit_Constant(self, node):
        self.found = True

    visit_Num = visit_Constant
    visit_Str = visit_Constant
    visit_Bytes = visit_Constant
    visit_NameConstant = visit_Constant


class Parser(NodeVisitor):
    """The main parser class.
    """

    def __init__(self, filename="", options=None, execution_context=None):
        # used in error messages:
        self.filename = filename
        # used to construct statement tree, also used for symbol table:
        self.state_stack = []
        # new statements are appended to this list:
        self.current_block = None
        self.current_context = None
        self.current_label = None
        # Allows temporarily overriding `current_process`:
        self._dummy_process = None
        self.errcnt = 0
        self.warncnt = 0
        self.program = execution_context if execution_context is not None \
                       else dast.Program() # Just in case
        self.cmdline_args = options
        self.module_args = Namespace()

    def get_option(self, option, default=None):
        if hasattr(self.cmdline_args, option):
            return getattr(self.cmdline_args, option)
        elif hasattr(self.module_args, option):
            return getattr(self.module_args, option)
        else:
            return default

    def push_state(self, node):
        self.state_stack.append((node,
                                 self.current_context,
                                 self.current_label,
                                 self.current_block))

    def pop_state(self):
        (_,
         self.current_context,
         self.current_label,
         self.current_block) = self.state_stack.pop()

    def is_in_setup(self):
        if self.current_process is None:
            return False
        elif isinstance(self.current_scope, dast.Function):
            return self.current_scope.name == "setup"

    def enter_query(self):
        pass

    def leave_query(self, audit=False):
        self.debug("Leaving query: " + str(self.current_query),
                   self.current_query)
        if audit and self.get_option('use_top_semantic', default=False):
            if self.current_parent is self.current_query:
                self.audit_query(self.current_parent)

    @property
    def current_parent(self):
        return self.state_stack[-1][0]

    @property
    def current_process(self):
        if self._dummy_process is not None:
            return self._dummy_process
        for node, _, _, _ in reversed(self.state_stack):
            if isinstance(node, dast.Process):
                return node
        return None

    @property
    def current_scope(self):
        for node, _, _, _ in reversed(self.state_stack):
            if isinstance(node, dast.NameScope):
                return node
        return None

    @property
    def current_query(self):
        for node, _, _, _ in reversed(self.state_stack):
            if isinstance(node, dast.QueryExpr):
                return node
            elif isinstance(node, dast.NameScope):
                # A query can not span a namescope (e.g. lambda expression):
                return None
        return None

    @property
    def current_query_scope(self):
        return self.current_query.scope \
            if self.current_query is not None else None

    @property
    def current_loop(self):
        for node, _, _, _ in reversed(self.state_stack):
            if isinstance(node, dast.ArgumentsContainer) or \
               isinstance(node, dast.ClassStmt):
                break
            elif isinstance(node, dast.LoopStmt):
                return node
        return None

    def visit_Module(self, node):
        dast.DistNode.reset_index()
        self.parse_module_header(node)
        self.program = dast.Program(None, node)
        self.program._compiler_options = self.module_args
        # Populate global scope with Python builtins:
        for name in PythonBuiltins:
            self.program.add_name(name)
        self.push_state(self.program)
        self.current_block = self.program.body
        self.current_context = Read(self.program)
        self.body(node.body)
        self.pop_state()

    def visit_Interactive(self, node):
        dast.DistNode.reset_index()
        self.program = dast.InteractiveProgram(None, node)
        # Populate global scope with Python builtins:
        for name in PythonBuiltins:
            self.program.add_name(name)
        self.push_state(self.program)
        contxtproc = dast.Process()
        self.push_state(contxtproc)


    # Helpers:

    def parse_module_header(self, node):
        assert isinstance(node, Module)

        if not (len(node.body) > 0 and
                isinstance(node.body[0], Expr) and
                isinstance(node.body[0].value, Str)):
            return

        self.debug("parsing module header...")
        opts = node.body[0].value.s.split()
        for o in opts:
            if o.startswith('--'):
                attr = o[2:].replace('-', '_')
                self.debug("setting module arg: " + attr, node.body[0])
                setattr(self.module_args, attr, True)
            else:
                self.debug("option not recognized: " + o, node.body[0])
        del node.body[0]

    def parse_bases(self, node):
        """Scans a ClassDef's bases list and checks whether the class defined by
           'node' is a DistProcess.

        A DistProcess is a class whose bases contain the name $KW_PROCESS_DEF.
        """
        isproc = False
        bases = []
        for b in node.bases:
            if (isinstance(b, Name) and b.id == KW_PROCESS_DEF):
                isproc = True
            else:
                bases.append(self.visit(b))
        return isproc, bases

    def parse_pattern_expr(self, node, literal=False):
        expr = self.create_expr(dast.PatternExpr, node)
        pp = PatternParser(self, literal)
        pattern = pp.visit(node)
        if pattern is None:
            self.error("invalid pattern", node)
            self.pop_state()
            return None
        expr.pattern = pattern
        self.pop_state()
        return expr

    def parse_decorators(self, node):
        assert hasattr(node, 'decorator_list')
        labels = set()
        notlabels = set()
        decorators = []
        for exp in node.decorator_list:
            if isinstance(exp, Call) and isinstance(exp.func, Name) and \
               exp.func.id == KW_DECORATOR_LABEL:
                for arg in exp.args:
                    l, negated = self.parse_label_spec(arg)
                    if negated:
                        notlabels |= l
                    else:
                        labels |= l
            else:
                decorators.append(self.visit(exp))
        return decorators, labels, notlabels

    def parse_label_spec(self, expr):
        negated = False
        if (type(expr) is UnaryOp and
                type(expr.operand) in {Set, Tuple, List}):
            names = expr.operand.elts
            negated = True
        elif type(expr) in {Set, Tuple, List}:
            names = expr.elts
        else:
            self.error("invalid label spec.", expr)
            names = []

        result = set()
        for elt in names:
            if type(elt) is not Name:
                self.error("invalid label spec.", elt)
            else:
                result.add(elt.id)
        return result, negated

    def parse_event_handler(self, node):
        if node.name == KW_RECV_EVENT:
            eventtype = dast.ReceivedEvent
        elif node.name == KW_SENT_EVENT:
            eventtype = dast.SentEvent
        else:
            # Impossible
            return None

        extras = []
        args = node.args
        if len(args.defaults) < len(args.args):
            extras.extend(args.args[:(len(args.args) - len(args.defaults))])
            args.args = args.args[(len(args.args) - len(args.defaults)):]
        if args.vararg:
            extras.append(args.vararg)
        if args.kwonlyargs:
            extras.extend(args.kwonlyargs)
        if args.kwarg:
            extras.append(args.kwarg)
        for node in extras:
            self.error("Malformed event spec: unknown argument {!r}.".
                       format(node.arg), node)

        events = []
        labels = set()
        notlabels = set()
        self.enter_query()
        for key, patexpr in zip(args.args, args.defaults):
            if key.arg == KW_EVENT_LABEL:
                ls, neg = self.parse_label_spec(patexpr)
                if neg:
                    notlabels |= ls
                else:
                    labels |= ls
                continue
            pat = self.parse_pattern_expr(patexpr)
            if key.arg == KW_MSG_PATTERN:
                events.append(dast.Event(self.current_process, ast=node,
                                         event_type=eventtype, pattern=pat))
                continue
            if len(events) == 0:
                self.error("invalid event spec: missing 'msg' argument.", node)
                # Add a phony event so we can recover as much as possible:
                events.append(dast.Event(self.current_process))
            if key.arg == KW_EVENT_SOURCE:
                events[-1].sources.append(pat)
            elif key.arg == KW_EVENT_DESTINATION:
                events[-1].destinations.append(pat)
            elif key.arg == KW_EVENT_TIMESTAMP:
                events[-1].timestamps.append(pat)
            else:
                self.warn("unrecognized event parameter '%s'" % key.arg, node)
        self.leave_query()
        return events, labels, notlabels

    def body(self, statements):
        """Process a block of statements.
        """
        for stmt in statements:
            self.current_context = None
            self.visit(stmt)
        if self.current_label is not None:
            # Create a noop statement to hold the last label:
            self.create_stmt(dast.NoopStmt, statements[-1], nopush=True)

    def proc_body(self, statements):
        """Process the body of a process definition.

        Process bodies differs from normal ClassDef bodies in that the names
        defined in this scope are visible to the whole process.
        """
        for stmt in statements:
            if (isinstance(stmt, FunctionDef) and stmt.name not in
                    {KW_RECV_EVENT, KW_SENT_EVENT}):
                self.debug("Adding function %s to process scope." % stmt.name,
                           stmt)
                self.current_scope.add_name(stmt.name)
            elif isinstance(stmt, ClassDef):
                self.debug("Adding class %s to process scope." % stmt.name,
                           stmt)
                self.current_scope.add_name(stmt.name)
            elif isinstance(stmt, Assign):
                for expr in stmt.targets:
                    if isinstance(expr, Name):
                        self.debug(
                            "Adding variable %s to process scope." % expr.id,
                            stmt)
                        self.current_scope.add_name(expr.id)
            elif isinstance(stmt, AugAssign):
                if isinstance(target, Name):
                    self.current_scope.add_name(target.id)
        for stmt in statements:
            self.visit(stmt)
        if self.current_label is not None:
            # Create a noop statement to hold the last label:
            self.create_stmt(dast.NoopStmt, statements[-1], nopush=True)

    def signature(self, node):
        """Process the argument lists."""
        assert isinstance(self.current_parent, dast.ArgumentsContainer)
        padding = len(node.args) - len(node.defaults)
        container = self.current_parent.args
        for arg in node.args[:padding]:
            if arg.annotation is not None:
                annotation = self.visit(arg.annotation)
                container.add_arg(arg.arg, annotation)
            else:
                container.add_arg(arg.arg)
        for arg, val in zip(node.args[padding:], node.defaults):
            if arg.annotation is not None:
                annotation = self.visit(arg.annotation)
                container.add_defaultarg(arg.arg, self.visit(val), annotation)
            else:
                container.add_defaultarg(arg.arg, self.visit(val))
        if node.vararg is not None:
            # Python 3.4 compatibility:
            if type(node.vararg) is str:
                container.add_vararg(node.vararg)
            else:
                if node.vararg.annotation is not None:
                    annotation = self.visit(node.vararg.annotation)
                    container.add_vararg(node.vararg.arg, annotation)
                else:
                    container.add_vararg(node.vararg.arg)
        if node.kwarg is not None:
            # Python 3.4 compatibility:
            if type(node.kwarg) is str:
                container.add_kwarg(node.kwarg)
            else:
                if node.kwarg.annotation is not None:
                    annotation = self.visit(node.kwarg.annotation)
                    container.add_kwarg(node.kwarg.arg, annotation)
                else:
                    container.add_kwarg(node.kwarg.arg)
        for kwarg, val in zip(node.kwonlyargs, node.kw_defaults):
            if val is not None:
                val = self.visit(val)
            if kwarg.annotation is not None:
                annotation = self.visit(kwarg.annotation)
                container.add_kwonlyarg(kwarg.arg, val, annotation)
            else:
                container.add_kwonlyarg(kwarg.arg, val)


    # Top-level blocks:

    def visit_ClassDef(self, node):
        isproc, bases = self.parse_bases(node)
        if isproc:
            if type(self.current_parent) is not dast.Program:
                self.error("Process definition must be at top level.", node)

            initfun = None
            bodyidx = None
            for idx, s in enumerate(node.body):
                if is_setup_func(s):
                    if initfun is None:
                        initfun = s
                        bodyidx = idx
                    else:
                        self.error("Duplicate setup() definition.", s)
            if initfun is None:
                self.warn("Process missing 'setup()' definition.", node)

            n = self.current_scope.add_name(node.name)
            proc = dast.Process(self.current_parent, node,
                                name=node.name, bases=bases)
            n.add_assignment(proc)
            proc.decorators, _, _ = self.parse_decorators(node)
            self.push_state(proc)
            self.program.processes.append(proc)
            self.program.body.append(proc)

            if initfun is not None:
                self.signature(initfun.args)
                if proc.args.kwarg is not None:
                    self.error("'setup()' can not have keyword argument.",
                               initfun)
            self.current_block = proc.body
            if bodyidx is not None:
                # setup() has to be parsed first:
                self.proc_body([node.body[bodyidx]] +
                               node.body[:bodyidx] + node.body[(bodyidx+1):])
            else:
                self.proc_body(node.body)
            dbgstr = ["Process ", proc.name, " has names: "]
            for n in proc._names.values():
                dbgstr.append("%s: %s; " % (n, str(n.get_typectx())))
            self.debug("".join(dbgstr))
            self.pop_state()
            if proc.entry_point is None:
                self.warn("Process %s missing '%s()' method." %
                          (proc.name, KW_PROCESS_ENTRY_POINT), node)

        else:
            clsobj = dast.ClassStmt(self.current_parent, node,
                                    name=node.name, bases=bases)
            if self.current_block is None or self.current_parent is None:
                self.error("Statement not allowed in this context.", ast)
            else:
                self.current_block.append(clsobj)
                n = self.current_scope.add_name(node.name)
                n.add_assignment(clsobj)
            clsobj.decorators, _, _ = self.parse_decorators(node)
            self.push_state(clsobj)
            self.current_block = clsobj.body
            self.body(node.body)
            dbgstr = ["Class ", clsobj.name, " has names: "]
            for n in clsobj._names.values():
                dbgstr.append("%s: %s; " % (n, str(n.get_typectx())))
            self.debug("".join(dbgstr))
            self.pop_state()

    def visit_FunctionDef(self, node):
        if (self.current_process is None or
                node.name not in {KW_SENT_EVENT, KW_RECV_EVENT}):
            # This is a normal method
            if self.current_parent is self.current_process:
                # Check for name conflict with internal methods:
                if node.name in InternalMethods:
                    self.error("Function name %r conflicts with DistAlgo "
                               "internals." % node.name, node)
            n = self.current_scope.add_name(node.name)
            s = self.create_stmt(dast.Function, node,
                                 params={"name" : node.name})
            n.add_assignment(s)
            s.process = self.current_process
            if isinstance(s.parent, dast.Process):
                if s.name == KW_PROCESS_ENTRY_POINT:
                    self.current_process.entry_point = s
                elif s.name == "setup":
                    self.current_process.setup = s
                else:
                    self.current_process.methods.append(s)
            elif (isinstance(s.parent, dast.Program) and
                  s.name == KW_ENTRY_POINT):
                # Create the node process:
                s.parent.nodecls = dast.Process(self.current_parent, None,
                                                name=NodeProcName)
                self._dummy_process = s.process = s.parent.nodecls
                self._dummy_process.entry_point = s
                # If we don't pop the entry_point from the module body then it
                # ends up getting printed twice:
                self.current_block.pop()
            # Ignore the label decorators:
            s.decorators, _, _ = self.parse_decorators(node)
            self.current_block = s.body
            if not self.is_in_setup():
                self.signature(node.args)
            self.body(node.body)
            dbgstr = [s.name, " has names: "]
            for n in s._names.values():
                dbgstr.append(("%s: %s; " % (n, str(n.get_typectx()))))
            self.debug("".join(dbgstr))
            self.pop_state()
            self._dummy_process = None

        else:
            # This is an event handler:
            h = dast.EventHandler(self.current_parent, node)
            # Parse decorators before adding h to node_stack, since decorators
            # should belong to the outer scope:
            h.decorators, h.labels, h.notlabels = self.parse_decorators(node)
            self.push_state(h)
            events, labels, notlabels = self.parse_event_handler(node)
            events = self.current_process.add_events(events)
            h.events = events
            h.labels |= labels
            h.notlabels |= notlabels
            if len(h.labels) == 0:
                h.labels = None
            if len(h.notlabels) == 0:
                h.notlabels = None
            for evt in events:
                evt.handlers.append(h)
                for v in evt.ordered_freevars:
                    if v is not None:
                        self.debug("adding event argument %s" % v)
                        h.args.add_arg(v.name)
            self.current_block = h.body
            self.body(node.body)
            self.pop_state()

    def check_await(self, node):
        if (isinstance(node, Call) and
            isinstance(node.func, Name) and
            node.func.id == KW_AWAIT):
            if len(node.args) <= 2:
                return True
            else:
                self.error("malformed await statement.", node)
                return None
        else:
            return False


    # Statements:
    #
    # The visit_* method for statements appends generated dast AST statements
    # to self.current_block.

    def create_stmt(self, stmtcls, ast, params=None, nopush=False):
        """Convenience method to instantiate a statement node and append to
        'current_block'.

        """
        if params is None:
            stmtobj = stmtcls(parent=self.current_parent, ast=ast)
        else:
            stmtobj = stmtcls(parent=self.current_parent, ast=ast, **params)
        stmtobj.label = self.current_label
        self.current_label = None

        if self.current_block is None or self.current_parent is None:
            self.error("Statement not allowed in this context.", ast)
        else:
            self.current_block.append(stmtobj)
        if not nopush:
            self.push_state(stmtobj)
        return stmtobj

    def create_expr(self, exprcls, ast, params=None, nopush=False):
        """Convenience method to instantiate an expression node.
        """
        if params is None:
            expr = exprcls(self.current_parent, ast=ast)
        else:
            expr = exprcls(self.current_parent, ast=ast, **params)
        if not nopush:
            self.push_state(expr)
        return expr


    def visit_Assign(self, node):
        stmtobj = self.create_stmt(dast.AssignmentStmt, node)
        self.current_context = Read(stmtobj)
        stmtobj.value = self.visit(node.value)
        self.current_context = Assignment(stmtobj, type=stmtobj.value)
        for target in node.targets:
            stmtobj.targets.append(self.visit(target))
        self.pop_state()

    def visit_AugAssign(self, node):
        stmtobj = self.create_stmt(dast.OpAssignmentStmt, node,
                                   params={'op':OperatorMap[type(node.op)]})
        self.current_context = Read(stmtobj)
        valexpr = self.visit(node.value)
        self.current_context = Assignment(stmtobj, type=valexpr)
        tgtexpr = self.visit(node.target)
        stmtobj.target = tgtexpr
        stmtobj.value = valexpr
        self.pop_state()

    def visit_ImportFrom(self, node):
        if hasattr(node, 'module'):
            stmtobj = self.create_stmt(dast.ImportFromStmt, node)
        else:
            stmtobj = self.create_stmt(dast.ImportStmt, node)
        for alias in node.names:
            if alias.asname is not None:
                nobj = self.current_scope.add_name(alias.asname)
                nalias = dast.Alias(stmtobj, alias,
                                    name=alias.name, asname=nobj)
            else:
                nobj = self.current_scope.add_name(alias.name)
                nalias = dast.Alias(stmtobj, alias, name=nobj)
            nobj.add_assignment(stmtobj)
            stmtobj.items.append(nalias)
        if hasattr(node, 'module'):
            stmtobj.module = node.module
            stmtobj.level = node.level
        self.pop_state()

    visit_Import = visit_ImportFrom

    def expr_check(self, names, minargs, maxargs, node,
                   keywords={}, optional_keywords={}):
        if not (isinstance(node, Call) and
                isinstance(node.func, Name) and
                (((isinstance(names, abc.Set) or
                   isinstance(names, abc.Mapping)) and
                  node.func.id in names) or
                 node.func.id == names)):
            return False

        try:
            if minargs is not None and len(node.args) < minargs:
                raise MalformedStatementError("too few arguments.")
            if maxargs is not None and len(node.args) > maxargs:
                raise MalformedStatementError("too many arguments.")
            if keywords is not None:
                for kw in node.keywords:
                    if kw.arg in keywords:
                        keywords -= {kw.arg}
                    elif optional_keywords is not None and \
                         kw.arg not in optional_keywords:
                        raise MalformedStatementError(
                            "unrecognized keyword: '%s'." % kw.arg)
                if len(keywords) > 0:
                    raise MalformedStatementError(
                        "missing required keywords: " + str(keywords))
        except MalformedStatementError as e:
            # Pre-format an error message for the common case:
            e.node = node
            e.name = node.func.id
            e.msg = "malformed {stname} statement: {msg}".format(
                stname=e.name, msg=e.reason)
            raise e

        return True

    def kw_check(self, names, node):
        if not isinstance(node, Name) or node.id not in names:
            return False
        return True

    def parse_message(self, node):
        if type(node) is Call and self.get_option('enable_object_pattern',
                                                  default=False):
            assert type(node.func) is Name
            expr = dast.TupleExpr(self.current_parent, node)
            elem = dast.ConstantExpr(self.current_parent, node.func)
            elem.value = node.func.id
            expr.subexprs.append(elem)
            elts = node.args
            for elt in elts:
                expr.subexprs.append(self.visit(elt))
            return expr
        else:
            return self.visit(node)

    def visit_Expr(self, node):
        l = extract_label(node.value)
        if l is not None and self.current_process is not None:
            self.current_label = l
            return
        stmtobj = None
        try:
            e = node.value
            if isinstance(self.current_parent, dast.Program) and \
               self.expr_check({KW_INC_VERB}, 1, 1, e):
                # Inc interface directive
                if isinstance(e.args[0], Str):
                    try:
                        self.current_parent.directives.extend(
                            parse(e.args[0].s).body)
                    except SyntaxError as err:
                        self.error("SyntaxError in %s directive: %s " %
                                   (KW_INC_VERB, err.msg), e.args[0])
                else:
                    self.error("%s directive argument must be string." %
                               KW_INC_VERB, e.args[0])

            elif self.expr_check(KW_AWAIT, 1, 2, e,
                                 keywords={},
                                 optional_keywords={KW_AWAIT_TIMEOUT}):
                stmtobj = self.create_stmt(dast.AwaitStmt, node)
                self.current_context = Read(stmtobj)
                branch = dast.Branch(stmtobj, node,
                                     condition=self.visit(e.args[0]))
                stmtobj.branches.append(branch)
                if len(e.args) == 2:
                    stmtobj.timeout = self.visit(e.args[1])
                if len(e.keywords) > 0:
                    if stmtobj.timeout is not None:
                        self.warn("duplicate timeout spec in await statement.",
                                  e)
                    stmtobj.timeout = self.visit(e.keywords[0].value)

            elif self.expr_check(KW_SEND, 1, 1, e, keywords={KW_SEND_TO},
                                 optional_keywords=None):
                stmtobj = self.create_stmt(dast.SimpleStmt, node)
                self.current_context = Read(stmtobj)
                stmtobj.expr = self.create_expr(dast.BuiltinCallExpr, e)
                self.current_context = Read(stmtobj.expr)
                stmtobj.expr.func = KW_SEND
                stmtobj.expr.args = [self.parse_message(e.args[0])]
                stmtobj.expr.keywords = [(kw.arg, self.visit(kw.value))
                                          for kw in e.keywords]
                self.pop_state()

            elif self.expr_check(KW_PRINT, 1, None, e,
                                 optional_keywords={KW_LEVEL, KW_SEP}):
                stmtobj = self.create_stmt(dast.OutputStmt, node)
                self.current_context = Read(stmtobj)
                stmtobj.message = [self.visit(arg) for arg in e.args]
                for kw in e.keywords:
                    if kw.arg == KW_LEVEL:
                        stmtobj.level = self.visit(kw.value)
                    elif kw.arg == KW_SEP:
                        stmtobj.separator = self.visit(kw.value)

            elif self.current_process is not None and \
                 self.expr_check(KW_RESET, 0, 1, e):
                stmtobj = self.create_stmt(dast.ResetStmt, node)
                self.current_context = Read(stmtobj)
                if len(e.args) > 0:
                    stmtobj.expr = self.visit(e.args[0])
                    if not isinstance(stmtobj.expr, dast.ConstantExpr):
                        self.error("Invalid argument in reset statement.", e)
                    elif stmtobj.expr.value not in ValidResetTypes:
                        self.error("Unknown argument in reset statement. "
                                   "Valid arguments are: " +
                                   str(ValidResetTypes), node)

            # Parse 'config' statements. These may appear at the module level or
            # the process level:
            elif self.expr_check(KW_CONFIG, 0, None, e, keywords=None):
                if isinstance(self.current_parent, dast.Process) or \
                   (isinstance(self.current_parent, dast.Function) and
                    (self.current_parent.name == KW_ENTRY_POINT or
                     self.current_parent.name == KW_PROCESS_ENTRY_POINT)):
                    self.current_process.configurations.extend(
                        self.parse_config_section(e))
                elif isinstance(self.current_parent, dast.Program):
                    self.current_parent.configurations.extend(
                        self.parse_config_section(e))
                else:
                    self.error("Invalid context for '%s' statement." %
                               KW_CONFIG, node)

            # 'yield' and 'yield from' should be statements, handle them here:
            elif type(e) is Yield:
                stmtobj = self.create_stmt(dast.YieldStmt, node)
                self.current_context = Read(stmtobj)
                stmtobj.value = self.visit(e.value)
            elif type(e) is YieldFrom:
                # 'yield' should be a statement, handle it here:
                stmtobj = self.create_stmt(dast.YieldFromStmt, node)
                self.current_context = Read(stmtobj)
                stmtobj.value = self.visit(e.value)

            else:
                stmtobj = self.create_stmt(dast.SimpleStmt, node)
                self.current_context = Read(stmtobj)
                stmtobj.expr = self.visit(node.value)

        except MalformedStatementError as e:
            self.error(e.msg, e.node)
        finally:
            if stmtobj is not None:
                self.pop_state()

    # ~~~

    def parse_branches_for_await(self, stmtobj, node):
        """Parse all branches except the first in an await statement.

        'stmtobj' is the DistNode instance of the await statement to be
        populated; 'node' is the input Python AST node corresponding to the
        outermost 'If' statement. This method is common code to parse the second
        to last branches of AwaitStmt and LoopingAwaitStmt, since the first
        branch has difference syntax for the two types of await ('if
        await(cond)...' for normal branching await, and just plain 'if cond...'
        for 'while await').

        """
        while True:
            else_ = node.orelse
            if len(else_) == 0:
                break
            elif len(else_) == 1 and isinstance(else_[0], If):
                # This is an 'elif' branch:
                node = else_[0]
                if self.expr_check(KW_AWAIT_TIMEOUT, 1 ,1, node.test):
                    stmtobj.timeout = self.visit(node.test.args[0])
                    if len(stmtobj.orelse) > 0:
                        self.warn("multiple timeout branches in await"
                                  " statement, all but the last one will be  "
                                  " ignored.", node)
                    if stmtobj.timeout is None:
                        self.warn("timeout branch in await statement without "
                                  "a timeout value.", node)
                        stmtobj.orelse = []
                    self.current_block = stmtobj.orelse
                    self.body(node.body)
                else:
                    branch = dast.Branch(stmtobj, node,
                                         condition=self.visit(node.test))
                    self.current_block = branch.body
                    self.body(node.body)
                    stmtobj.branches.append(branch)
            else:
                # This is a dangling 'else' branch, treat it as the timeout
                # branch:
                if stmtobj.timeout is None:
                    self.warn("timeout branch in await statement without "
                              "a timeout value.", node)
                if len(stmtobj.orelse) > 0:
                    self.warn("multiple timeout branches in await"
                              " statement, all but the last one will be  "
                              " ignored.", node)
                    stmtobj.orelse = []
                self.current_block = stmtobj.orelse
                self.body(else_)
                break

    def visit_If(self, node):
        stmtobj = None
        try:
            if self.expr_check(KW_AWAIT, 1, 1, node.test):
                stmtobj = self.create_stmt(dast.AwaitStmt, node)
                self.current_context = Read(stmtobj)
                branch = dast.Branch(stmtobj, node.test,
                                     condition=self.visit(node.test.args[0]))
                if len(node.test.args) == 2:
                    stmtobj.timeout = self.visit(node.test.args[1])
                if len(node.test.keywords) > 0:
                    if stmtobj.timeout is not None:
                        self.warn("duplicate timeout spec in await statement.",
                                  e)
                    stmtobj.timeout = self.visit(e.keywords[0].value)
                self.current_block = branch.body
                self.body(node.body)
                stmtobj.branches.append(branch)
                self.parse_branches_for_await(stmtobj, node)

            else:
                stmtobj = self.create_stmt(dast.IfStmt, node)
                self.current_context = Read(stmtobj)
                stmtobj.condition = self.visit(node.test)
                self.current_block = stmtobj.body
                self.body(node.body)
                self.current_block = stmtobj.elsebody
                self.body(node.orelse)

        except MalformedStatementError as e:
            self.error(e.msg, e.node)
        finally:
            if stmtobj is not None:
                self.pop_state()

    def visit_For(self, node):
        s = self.create_stmt(dast.ForStmt, node)
        try:
            self.current_context = Assignment(s)
            s.domain = self.parse_domain_spec(node)
            self.current_context = None
            self.current_block = s.body
            self.body(node.body)
            self.current_block = s.elsebody
            self.body(node.orelse)
        except MalformedStatementError as e:
            self.error(e.msg if e.msg is not None else e.reason,
                       e.node)
        finally:
            self.pop_state()

    def visit_While(self, node):
        s = None
        try:
            if self.kw_check({KW_AWAIT}, node.test):
                whilenode = node
                s = self.create_stmt(dast.LoopingAwaitStmt, node)
                self.current_context = Read(s)
                if len(node.body) != 1 or not isinstance(node.body[0], If):
                    self.error("malformed 'while await' statement.", node)
                if isinstance(node.body[0], If):
                    node = node.body[0]
                    branch = dast.Branch(s, node,
                                         condition=self.visit(node.test))
                    self.current_block = branch.body
                    self.body(node.body)
                    s.branches.append(branch)
                    self.parse_branches_for_await(s, node)
                self.current_block = s.orfail
                self.body(whilenode.orelse)

            else:
                s = self.create_stmt(dast.WhileStmt, node)
                self.current_context = Read(s)
                s.condition = self.visit(node.test)
                self.current_block = s.body
                self.body(node.body)
                self.current_block = s.elsebody
                self.body(node.orelse)

        except MalformedStatementError as e:
            self.error(e.msg, e.node)
        finally:
            if s is not None:
                self.pop_state()

    def visit_With(self, node):
        s = self.create_stmt(dast.WithStmt, node)
        for item in node.items:
            self.current_context = Read(s)
            ctxexpr = self.visit(item.context_expr)
            if item.optional_vars is not None:
                self.current_context = Assignment(s, type=ctxexpr)
                s.items.append((ctxexpr, self.visit(item.optional_vars)))
            else:
                s.items.append((ctxexpr, None))
        self.current_context = None
        self.current_block = s.body
        self.body(node.body)
        self.pop_state()

    def visit_Pass(self, node):
        self.create_stmt(dast.PassStmt, node, nopush=True)

    def visit_Break(self, node):
        loop = self.current_loop
        if loop is None:
            self.warn("Possible use of 'break' outside loop.", node)
        self.create_stmt(dast.BreakStmt, node, nopush=True,
                         params={"loopstmt": loop})

    def visit_Continue(self, node):
        loop = self.current_loop
        if loop is None:
            self.warn("Possible use of 'continue' outside loop.", node)
        self.create_stmt(dast.ContinueStmt, node, nopush=True,
                         params={"loopstmt": loop})

    def visit_Delete(self, node):
        s = self.create_stmt(dast.DeleteStmt, node)
        self.current_context = Delete(s)
        for target in node.targets:
            s.targets.append(self.visit(target))
        self.pop_state()

    def visit_Try(self, node):
        s = self.create_stmt(dast.TryStmt, node)
        self.current_block = s.body
        self.body(node.body)
        self.current_context = Read(s)
        for handler in node.handlers:
            h = dast.ExceptHandler(s, handler)
            h.name = handler.name
            if h.name is not None:
                n = self.current_scope.find_name(h.name)
                if n is None:
                    self.current_scope.add_name(h.name)
                n.add_assignment(s)
            if handler.type is not None:
                h.type = self.visit(handler.type)
            self.current_block = h.body
            self.body(handler.body)
            s.excepthandlers.append(h)
        self.current_block = s.elsebody
        self.body(node.orelse)
        self.current_block = s.finalbody
        self.body(node.finalbody)
        self.pop_state()

    def visit_Assert(self, node):
        s = self.create_stmt(dast.AssertStmt, node)
        self.current_context = Read(s)
        s.expr = self.visit(node.test)
        if node.msg is not None:
            s.msg = self.visit(node.msg)
        self.pop_state()

    def visit_Global(self, node):
        if self.current_process is not None:
            self.warn("'global' statement inside process is redundant and "
                      "ignored.", node)
        else:
            self.create_stmt(dast.GlobalStmt, node,
                             {"names": list(node.names)})
            for name in node.names:
                localname = self.current_scope.find_name(name, local=True)
                if localname is not None:
                    self.warn("name '%s' used before declared 'global'." %
                              name, node)
                nobj = self.program.find_name(name)
                if nobj is None:
                    nobj = self.program.add_name(name)
                self.debug("Linking global name '%s'" % name)
                self.current_scope.link_name(nobj)
            self.pop_state()

    def visit_Nonlocal(self, node):
        self.create_stmt(dast.NonlocalStmt, node, {"names": list(node.names)})
        if self.current_scope.parent_scope is None:
            self.error("No nonlocal scope found.", node)
        else:
            for name in node.names:
                nobj = self.current_scope.find_name(name, local=True)
                if nobj is not None:
                    self.warn("Variable '%s' used before declared 'nonlocal'." %
                              name, node)
                nobj = self.current_scope.parent_scope.find_name(name, local=False)
                if nobj is None:
                    self.warn("Unable to determine scope for nonlocal var %s" %
                              name, node)
                else:
                    self.debug("Linking nonlocal name '%s'" % name)
                    self.current_scope.link_name(nobj)
        self.pop_state()

    def visit_Return(self, node):
        s = self.create_stmt(dast.ReturnStmt, node)
        self.current_context = Read(s)
        if node.value is not None:
            s.value = self.visit(node.value)
        self.pop_state()

    def visit_Raise(self, node):
        s = self.create_stmt(dast.RaiseStmt, node)
        self.current_context = Read(s)
        if node.exc is not None:
            s.expr = self.visit(node.exc)
        if node.cause is not None:
            s.cause = self.visit(node.cause)
        self.pop_state()


    # Expressions:
    #
    # The visit_* methods for expressions return the newly
    # constructed dast AST node

    def visit_Attribute(self, node):
        expr = self.create_expr(dast.AttributeExpr, node)
        if (isinstance(self.current_context, FunCall) and
                node.attr in KnownUpdateMethods):
            # Calling a method that is known to update an object's state is an
            # Update operation:
            self.current_context = Update(self.current_context.node,
                                          self.current_context.type)
        elif isinstance(self.current_context, Assignment) or\
             isinstance(self.current_context, Delete):
            # Assigning to or deleting an attribute of an object updates that
            # object:
            self.current_context = Update(self.current_context.node,
                                          self.current_context.type)
        expr.value = self.visit(node.value)
        expr.attr = node.attr
        self.pop_state()
        if isinstance(expr.value, dast.SelfExpr):
            if node.attr == 'id':
                self.warn("'self.id' is deprecated, use 'self' instead!", node)
                return expr.value
            # Need to update the namedvar object
            n = self.current_process.find_name(expr.attr)
            expr = self.create_expr(dast.SimpleExpr, node)
            if n is None:
                if (self.is_in_setup() and
                        isinstance(self.current_context, Assignment)):
                    self.debug("Adding name '%s' to process scope"
                               " from setup()." % node.attr, node)
                    n = self.current_process.add_name(node.attr)
                    n.add_assignment(self.current_context.node,
                                     self.current_context.type)
                    n.set_scope(self.current_process)
                else:
                    self.error("Undefined process state variable: " +
                               str(node.attr), node)
            else:
                if isinstance(self.current_context, Assignment) or \
                   isinstance(self.current_context, Delete):
                    self.debug("Assignment to variable '%s'" % str(n), node)
                    n.add_assignment(self.current_context.node,
                                     self.current_context.type)
                elif isinstance(self.current_context, Update):
                    self.debug("Update to process variable '%s'" % str(n), node)
                    n.add_update(self.current_context.node,
                                 self.current_context.type)
                else:
                    n.add_read(expr)
            expr.value = n
            self.pop_state()
        return expr

    def ensure_one_arg(self, name, node):
        l = len(node.args)
        if l != 1:
            self.error("'%s' takes exactly one argument (%d given)" % (name, l),
                       node)
            return False
        return True

    def ensure_sequence_arg(self, name, node):
        l = len(node.args)
        if l > 1:
            self.error("'%s' takes zero or one argument (%d given)" % (name, l),
                       node)
            return False
        if l == 1 and not hasattr(node.args[0], "elts"):
            return False
        return True

    def parse_event_expr(self, node, literal=False):
        if sys.version_info < (3, 5):
            if (node.starargs is not None or node.kwargs is not None):
                self.warn("extraneous arguments in event expression.", node)
        pattern = self.parse_pattern_expr(node.args[0], literal)
        if node.func.id == KW_RECV_QUERY:
            event = dast.Event(self.current_process,
                               event_type=dast.ReceivedEvent,
                               pattern=pattern)
        elif node.func.id == KW_SENT_QUERY:
            event = dast.Event(self.current_process,
                               event_type=dast.SentEvent,
                               pattern=pattern)
        else:
            self.error("unknown event specifier", node)
            return None
        for kw in node.keywords:
            pat = self.parse_pattern_expr(kw.value, literal)
            if kw.arg == KW_EVENT_SOURCE:
                event.sources.append(pat)
            elif kw.arg == KW_EVENT_DESTINATION:
                event.destinations.append(pat)
            elif kw.arg == KW_EVENT_TIMESTAMP:
                event.timestamps.append(pat)
            else:
                self.warn("unknown keyword in query.", node)
        return self.current_process.add_event(event)

    def event_from_pattern(self, node, event_type):
        assert isinstance(node, dast.PatternExpr)
        pattern = node.pattern
        event = dast.Event(self.current_process,
                           event_type=event_type)
        if not isinstance(pattern, dast.TuplePattern):
            self.error("malformed event pattern.", node)
        elif self.get_option('full_event_pattern', default=False):
            if len(pattern.value) != 3:
                self.error("malformed event pattern.", node)
            else:
                event.pattern = dast.PatternExpr(node.parent,
                                                 pattern=pattern.value[2])
                envpat = pattern.value[1]
                if isinstance(envpat, dast.TuplePattern):
                    if len(envpat.value) != 3:
                        self.warn("possible malformed envelope pattern.", node)
                    else:
                        event.timestamps.append(
                            dast.PatternExpr(node.parent,
                                             pattern=envpat.value[0]))
                        event.destinations.append(
                            dast.PatternExpr(node.parent,
                                             pattern=envpat.value[1]))
                        event.sources.append(
                            dast.PatternExpr(node.parent,
                                             pattern=envpat.value[2]))
        else:
            if len(pattern.value) != 2:
                self.error("malformed event pattern.", node)
            else:
                event.pattern = dast.PatternExpr(node.parent,
                                                 pattern=pattern.value[0])
                event.sources.append(
                    dast.PatternExpr(node.parent, pattern=pattern.value[1]))
        return self.current_process.add_event(event)

    def pattern_from_event(self, node, literal=False):
        if not isinstance(node, dast.Event):
            return None
        expr = self.create_expr(dast.PatternExpr if not literal else
                                dast.LiteralPatternExpr,
                                node.ast)
        pattern = dast.TuplePattern(node.parent)

        # Pattern structure:
        # (TYPE, ENVELOPE, MESSAGE)
        # ENVELOPE: (TIMESTAMP, DESTINATION, SOURCE)
        if isinstance(node.type, dast.EventType):
            pattern.value.append(
                dast.ConstantPattern(
                    pattern,
                    value=self.current_scope.add_name(
                        node.type.__name__)))
        else:
            pattern.value.append(dast.FreePattern(pattern))

        env = dast.TuplePattern(pattern)
        if (len(node.timestamps) == 0):
            env.value.append(dast.FreePattern(env))
        elif len(node.timestamps) == 1:
            env.value.append(node.timestamps[0].pattern.clone())
            env.value[-1]._parent = env
        else:
            self.error("multiple timestamp spec in event pattern.", node)
        if (len(node.destinations) == 0):
            env.value.append(dast.FreePattern(env))
        elif len(node.destinations) == 1:
            env.value.append(node.destinations[0].pattern.clone())
            env.value[-1]._parent = env
        else:
            self.error("multiple destination spec in event pattern.", node)
        if (len(node.sources) == 0):
            env.value.append(dast.FreePattern(env))
        elif len(node.sources) == 1:
            env.value.append(node.sources[0].pattern.clone())
            env.value[-1]._parent = env
        else:
            self.error("multiple source spec in event pattern.", node)

        pattern.value.append(env)
        if node.pattern is None:
            msgpat = dast.FreePattern(pattern)
        else:
            msgpat = node.pattern.pattern.clone()
            msgpat._parent = pattern
        pattern.value.append(msgpat)
        expr.pattern = pattern
        self.pop_state()
        return expr

    def test_domain_spec(self, node):
        try:
            if (self.current_process is not None and
                    isinstance(node, Call) and
                    self.expr_check({KW_RECV_QUERY, KW_SENT_QUERY},
                                    1, 1, node,
                                    optional_keywords=EventKeywords)):
                return True
            elif (isinstance(node, Compare) and len(node.ops) == 1 and
                  type(node.ops[0]) is In):
                return True
            elif isinstance(node, comprehension) or isinstance(node, For):
                return True
        except MalformedStatementError as e:
                pass
        return False

    def parse_domain_spec(self, node):
        if (self.current_process is not None and
                isinstance(node, Call) and
                self.expr_check({KW_RECV_QUERY, KW_SENT_QUERY},
                                1, 1, node,
                                optional_keywords=EventKeywords)):
            # As a short hand, "sent" and "rcvd" can be used as a domain
            # spec: some(rcvd(EVENT_PATTERN) | PRED) is semantically
            # equivalent to some(EVENT_PATTERN in rcvd | PRED).
            expr = self.create_expr(dast.DomainSpec, node)
            event = self.parse_event_expr(node, literal=False)
            if event is not None:
                event.record_history = True
                expr.pattern = self.pattern_from_event(event)
                if node.func.id == KW_RECV_QUERY:
                    expr.domain = self.create_expr(dast.ReceivedExpr, node)
                else:
                    expr.domain = self.create_expr(dast.SentExpr, node)
                expr.domain.event = event
                self.pop_state()
            self.pop_state()
            return expr
        elif (isinstance(node, Compare) and len(node.ops) == 1 and
              type(node.ops[0]) is In):
            expr = self.create_expr(dast.DomainSpec, node)
            self.current_context = Assignment(expr)
            expr.pattern = self.parse_pattern_expr(node.left)
            self.current_context = IterRead(expr, type=expr.pattern)
            expr.domain = self.visit(node.comparators[0])
            if isinstance(expr.domain, dast.HistoryExpr):
                expr.pattern = self.pattern_from_event(expr.domain.event)
            self.pop_state()
            return expr
        elif isinstance(node, comprehension) or isinstance(node, For):
            expr = self.create_expr(dast.DomainSpec, node)
            self.current_context = Assignment(expr)
            if self.get_option('enable_iterator_pattern', default=False):
                expr.pattern = self.parse_pattern_expr(node.target)
            else:
                expr.pattern = self.visit(node.target)
            self.current_context = IterRead(expr, type=expr.pattern)
            expr.domain = self.visit(node.iter)
            if isinstance(expr.domain, dast.HistoryExpr):
                expr.pattern = self.pattern_from_event(expr.domain.event)
            self.pop_state()
            return expr
        else:
            raise MalformedStatementError("malformed domain specifier.", node)

    def parse_quantified_expr(self, node):
        if node.func.id == KW_EXISTENTIAL_QUANT:
            context = dast.ExistentialOp
        elif node.func.id == KW_UNIVERSAL_QUANT:
            context = dast.UniversalOp
        else:
            raise MalformedStatementError("unknown quantifier.", node,
                                          node.func.id)

        expr = self.create_expr(dast.QuantifiedExpr, node, {'op': context})
        self.current_context = Read(expr)
        self.enter_query()
        try:
            expr.domains, predicates = self.parse_domains_and_predicate(node)
            if len(predicates) > 1:
                self.warn("multiple predicates in quantified expression, "
                          "first one is used, the rest are ignored.", node)
            expr.predicate = predicates[0]
        finally:
            self.leave_query(audit=True)
            self.pop_state()
        return expr

    def parse_comprehension(self, node):
        if node.func.id == KW_COMP_SET:
            expr_type = dast.SetCompExpr
        elif node.func.id == KW_COMP_LIST:
            expr_type = dast.ListCompExpr
        elif node.func.id == KW_COMP_DICT:
            expr_type = dast.DictCompExpr
        elif node.func.id == KW_COMP_TUPLE:
            expr_type = dast.TupleCompExpr

        expr = self.create_expr(expr_type, node)
        self.enter_query()

        first_arg = node.args[0]
        for arg in node.args[1:]:
            condition = None
            # try to treat it as domain spec first:
            if self.test_domain_spec(arg):
                try:
                    dom = self.parse_domain_spec(arg)
                    if len(dom.freevars) == 0:
                        # no freevars, degenerate to membership test:
                        condition = self.create_expr(dast.ComparisonExpr, arg)
                        condition.left = Pattern2Constant(
                            self.current_parent).visit(dom.pattern.pattern)
                        if condition.left is None:
                            self.error("internal error: unable to generate "
                                       "constant from pattern.", node)
                        condition.comparator = dast.InOp
                        condition.right = dom.domain
                        self.pop_state()
                    else:
                        condition = dom
                except MalformedStatementError as e:
                    self.error("malformed domain spec: " + e.reason, e.node)
            else:
                # if not, then it's just a boolean condition:
                condition = self.visit(arg)
            if condition is not None:
                expr.conditions.append(condition)

        if expr_type is dast.DictCompExpr:
            if not (isinstance(first_arg, Tuple) and
                    len(first_arg.elts) == 2):
                self.error("malformed element in dict comprehension.",
                           first_arg)
            else:
                kv = dast.KeyValue(expr)
                kv.key = self.visit(node.key)
                kv.value = self.visit(node.value)
                expr.elem = kv
        else:
            expr.elem = self.visit(first_arg)

        self.leave_query()
        self.pop_state()
        return expr

    def audit_query(self, expr):
        self.debug("auditing " + str(expr), expr)
        self.debug("...freevars: " + str(expr.freevars), expr)
        self.debug("...boundvars: " + str(expr.boundvars), expr)
        intersect = {v.name for v in expr.ordered_freevars} & \
                    {v.name for v in expr.ordered_boundvars}
        if intersect:
            msg = ("query variables " +
                   " ".join(["'" + n + "'" for n in intersect]) +
                   " are both free and bound.")
            self.error(msg, expr)

    def parse_aggregates(self, node):
        if node.func.id == KW_AGGREGATE_SUM:
            expr_type = dast.SumExpr
        elif node.func.id == KW_AGGREGATE_SIZE:
            expr_type = dast.SizeExpr
        elif node.func.id == KW_AGGREGATE_MIN:
            expr_type = dast.MinExpr
        elif node.func.id == KW_AGGREGATE_MAX:
            expr_type = dast.MaxExpr

        expr = self.create_expr(expr_type, node)
        self.current_context = Read(expr)
        first_arg = node.args[0]
        node.args = node.args[1:]
        try:
            expr.domains, expr.conditions = self.parse_domains_and_predicate(node)
            expr.elem = self.visit(first_arg)
        finally:
            self.pop_state()
        return expr

    def parse_domains_and_predicate(self, node):
        preds = []
        # Find predicate:
        for kw in node.keywords:
            if kw.arg == KW_SUCH_THAT:
                preds.append(kw.value)
            else:
                self.error("Unknown keyword '%s' in comprehension expression." %
                           kw.arg, node)
        # ..if no predicate found, then default to True:
        if len(preds) == 0:
            preds= [NameConstant(True)]
        domains = node.args
        if len(domains) == 0:
            self.warn("No domain specifiers in comprehension expression.", node)
        dadomains = [self.parse_domain_spec(node) for node in domains]
        dapredicates = [self.visit(pred) for pred in preds]
        return dadomains, dapredicates

    class NameTransformer(NodeTransformer):
        def visit_Name(self, node):
            return Str(node.id)

    def parse_config_value(self, vnode):
        value = None
        # Configuration values can not contain variables, so we treat all
        # 'Name's as 'Str's:
        vnode = Parser.NameTransformer().visit(vnode)
        if isinstance(vnode, Num) or isinstance(vnode, Str) or \
           isinstance(vnode, Bytes) or isinstance(vnode, NameConstant) or \
           isinstance(vnode, Set) or isinstance(vnode, List) or \
           isinstance(vnode, Tuple):
            value = self.visit(vnode)
        else:
            self.error("Invalid configuration value.", vnode)
        return value

    def parse_config_section(self, node):
        res = []
        for argexpr in node.args:
            if (isinstance(argexpr, Compare) and
                  isinstance(argexpr.left, Name) and
                  len(argexpr.ops) == 1 and
                  type(argexpr.ops[0]) is Is):
                key = argexpr.left.id.casefold()
                value = self.parse_config_value(argexpr.comparators[0])
                if value is not None:
                    res.append((key, value))
        for kw in node.keywords:
            key = kw.arg.casefold()
            vnode = kw.value
            value = self.parse_config_value(vnode)
            if value is not None:
                res.append((key, value))
        return res

    def visit_Call(self, node):
        try:
            if self.expr_check(Quantifiers, 1, None, node,
                               optional_keywords={KW_SUCH_THAT}):
                return self.parse_quantified_expr(node)

            if self.expr_check(ComprehensionTypes, 2, None, node):
                return self.parse_comprehension(node)

            if self.current_process is not None and \
               self.expr_check({KW_RECV_QUERY, KW_SENT_QUERY}, 1, 1, node,
                               optional_keywords=EventKeywords):
                if isinstance(self.current_context, IterRead):
                    if node.func.id == KW_RECV_QUERY:
                        expr = self.create_expr(dast.ReceivedExpr, node)
                    else:
                        expr = self.create_expr(dast.SentExpr, node)
                    expr.context = self.current_context.type
                    expr.event = self.parse_event_expr(
                        node, literal=(not self.get_option(
                            'enable_iterator_pattern', default=False)))
                    self.pop_state()
                    if expr.event is not None:
                        expr.event.record_history = True
                    return expr
                else:
                    outer = self.create_expr(dast.ComparisonExpr, node)
                    outer.comparator = dast.InOp
                    if node.func.id == KW_RECV_QUERY:
                        expr = self.create_expr(dast.ReceivedExpr, node)
                    else:
                        expr = self.create_expr(dast.SentExpr, node)
                    if self.current_context is not None:
                        expr.context = self.current_context.type
                    event = self.parse_event_expr(
                        node, literal=(not self.get_option(
                            'enable_membertest_pattern', default=False)))
                    self.pop_state()
                    expr.event = event
                    outer.right = expr
                    if event is not None:
                        outer.left = self.pattern_from_event(
                            event, literal=(not self.get_option(
                                'enable_membertest_pattern', default=False)))
                        event.record_history = True
                    self.pop_state()
                    return outer
        except MalformedStatementError as e:
            self.error("malformed {name} expression: {reason}".format(
                name=e.name if e.name is not None else "",
                reason=e.reason), e.node)
            return dast.SimpleExpr(self.current_parent)

        expr = None
        if self.expr_check(ApiMethods, None, None, node,
                           keywords=None, optional_keywords=None):
            self.debug("Api method call: " + node.func.id, node)
            expr = self.create_expr(dast.ApiCallExpr, node)
            expr.func = node.func.id
        elif self.expr_check(BuiltinMethods, None, None, node,
                             keywords=None, optional_keywords=None):
            self.debug("Builtin method call: " + node.func.id, node)
            expr = self.create_expr(dast.BuiltinCallExpr, node)
            expr.func = node.func.id
        elif self.expr_check({KW_SETUP}, None, None, node,
                             keywords=None, optional_keywords=None):
            self.debug("Setup expression. ", node)
            expr = self.create_expr(dast.SetupExpr, node)
        elif self.expr_check({KW_START}, None, None, node,
                             keywords=None, optional_keywords=None):
            self.debug("Start expression. ", node)
            expr = self.create_expr(dast.StartExpr, node)
        elif self.expr_check({KW_CONFIG}, None, None, node,
                             keywords=None, optional_keywords=None):
            self.debug("Config expression. ", node)
            expr = self.create_expr(dast.ConfigExpr, node)
        elif self.expr_check(AggregateMap, 1, None, node,
                             keywords={}, optional_keywords={}):
            self.debug("Aggregate: " + node.func.id, node)
            expr = self.create_expr(AggregateMap[node.func.id], node)
        else:
            if isinstance(node.func, Name):
                self.debug("Method call: " + str(node.func.id), node)
            expr = self.create_expr(dast.CallExpr, node)
            self.current_context = FunCall(expr)
            expr.func = self.visit(node.func)

        self.current_context = Read(expr)
        expr.args = [self.visit(a) for a in node.args]
        expr.keywords = [(kw.arg, self.visit(kw.value))
                         for kw in node.keywords]
        # Python 3.5 got rid of `starargs' and `kwargs' on Call objects:
        if sys.version_info < (3 ,5):
            expr.starargs = self.visit(node.starargs) \
                            if node.starargs is not None else None
            expr.kwargs = self.visit(node.kwargs) \
                          if node.kwargs is not None else None
        self.pop_state()
        return expr

    def visit_Name(self, node):
        if node.id in {KW_TRUE, KW_FALSE, KW_NULL}:
            if type(self.current_context) in {Assignment, Update, Delete}:
                self.warn("Constant expression in update context.", node)

            if node.id == KW_TRUE:
                return self.create_expr(dast.TrueExpr, node, nopush=True)
            elif node.id == KW_FALSE:
                return self.create_expr(dast.FalseExpr, node, nopush=True)
            elif node.id == KW_NULL:
                return self.create_expr(dast.NoneExpr, node, nopush=True)
        if self.current_process is not None and node.id == KW_SELF:
            return self.create_expr(dast.SelfExpr, node, nopush=True)

        if (self.current_process is not None and
                (node.id in {KW_RECV_QUERY, KW_SENT_QUERY})):
            if node.id == KW_RECV_QUERY:
                expr = self.create_expr(dast.ReceivedExpr, node)
                event_type = dast.ReceivedEvent
            else:
                expr = self.create_expr(dast.SentExpr, node)
                event_type = dast.SentEvent

            if (isinstance(self.current_context, Read) and
                    isinstance(self.current_context.type, dast.PatternExpr)):
                expr.context = self.current_context.type
                event = self.event_from_pattern(expr.context, event_type)
                expr.event = event
                event.record_history = True
            else:
                self.error("Invalid context for '%s'" % node.id, node)
            self.pop_state()
            return expr

        # NamedVar is not by itself an Expression, we'll have to wrap it in a
        # SimpleExpr:
        expr = self.create_expr(dast.SimpleExpr, node)
        n = self.current_scope.find_name(node.id, local=False)
        if isinstance(self.current_context, Assignment) or\
           isinstance(self.current_context, Delete):
            if n is None:
                self.debug("Adding name %s to %s" % (node.id,
                                                     self.current_scope), node)
                n = self.current_scope.add_name(node.id)
            n.add_assignment(self.current_context.node,
                             self.current_context.type)
        elif isinstance(self.current_context, Update):
            if n is None:
                self.warn("Possible use of uninitialized variable '%s'" %
                          node.id, node)
                self.debug(str(self.current_scope.parent_scope), node)
                n = self.current_scope.add_name(node.id)
            n.add_update(self.current_context.node, None,
                         self.current_context.type)
        elif isinstance(self.current_context, Read) or \
             isinstance(self.current_context, FunCall):
            if n is None:
                self.warn("Possible use of uninitialized variable '%s'" %
                          node.id, node)
                self.debug(str(self.current_scope.parent_scope), node)
                if self.current_scope.parent_scope is not None:
                    self.debug(self.current_scope.parent_scope._names, node)
                else:
                    self.debug(self.current_scope._names, node)
                n = self.current_scope.add_name(node.id)
            n.add_read(expr)
        if n is None:
            # A fallback for the cases we don't care about (i.e.
            # annontations)
            n = self.current_scope.add_name(node.id)
        expr.value = n
        self.pop_state()
        return expr

    def visit_Str(self, node):
        expr = self.create_expr(dast.ConstantExpr, node)
        expr.value = node.s
        self.pop_state()
        return expr

    def visit_Bytes(self, node):
        expr = self.create_expr(dast.ConstantExpr, node)
        expr.value = node.s
        self.pop_state()
        return expr

    def visit_Num(self, node):
        expr = self.create_expr(dast.ConstantExpr, node)
        expr.value = node.n
        self.pop_state()
        return expr

    # Since Python 3.4:
    def visit_NameConstant(self, node):
        if node.value == True:
            return self.create_expr(dast.TrueExpr, node, nopush=True)
        elif node.value == False:
            return self.create_expr(dast.FalseExpr, node, nopush=True)
        elif node.value == None:
            return self.create_expr(dast.NoneExpr, node, nopush=True)
        else:
            raise NotImplementedError("Unrecognized NameConstant %s." % repr(node.value))

    def visit_Tuple(self, node):
        expr = self.create_expr(dast.TupleExpr, node)
        for item in node.elts:
            expr.subexprs.append(self.visit(item))
        self.pop_state()
        return expr

    def visit_List(self, node):
        expr = self.create_expr(dast.ListExpr, node)
        for item in node.elts:
            expr.subexprs.append(self.visit(item))
        self.pop_state()
        return expr

    def visit_Set(self, node):
        expr = self.create_expr(dast.SetExpr, node)
        for item in node.elts:
            expr.subexprs.append(self.visit(item))
        self.pop_state()
        return expr

    def visit_Dict(self, node):
        expr = self.create_expr(dast.DictExpr, node)
        for key in node.keys:
            expr.keys.append(self.visit(key))
        for value in node.values:
            expr.values.append(self.visit(value))
        self.pop_state()
        return expr

    def visit_BinOp(self, node):
        e = self.create_expr(dast.BinaryExpr, node,
                             {"op": OperatorMap[type(node.op)]})
        e.left = self.visit(node.left)
        e.right = self.visit(node.right)
        self.pop_state()
        return e

    def visit_BoolOp(self, node):
        e = self.create_expr(dast.LogicalExpr, node,
                             {"op": OperatorMap[type(node.op)]})
        for v in node.values:
            e.subexprs.append(self.visit(v))
        self.pop_state()
        return e

    def visit_Compare(self, node):
        if len(node.ops) > 1:
            self.error("Explicit parenthesis required in comparison expression",
                       node)
            return None
        outer = None
        # We make all negation explicit:
        if type(node.ops[0]) in NegatedOperators:
            outer = self.create_expr(dast.LogicalExpr, node)
            outer.operator = dast.NotOp

        expr = self.create_expr(dast.ComparisonExpr, node)

        if self.get_option('enable_membertest_pattern', default=False):
            # DistAlgo: overload "in" to allow pattern matching
            if isinstance(node.ops[0], In) or \
                   isinstance(node.ops[0], NotIn):
                # Backward compatibility: only assume pattern if containing free
                # var
                pf = PatternFinder()
                pf.visit(node.left)
                if pf.found:
                    expr.left = self.parse_pattern_expr(node.left)
        if expr.left is None:
            expr.left = self.visit(node.left)
        self.current_context = Read(expr.left)
        expr.right = self.visit(node.comparators[0])
        if (isinstance(expr.right, dast.HistoryExpr) and
                expr.right.event is not None):
            # Must replace short pattern format with full pattern here:
            expr.left = self.pattern_from_event(expr.right.event)

        if outer is not None:
            expr.comparator = NegatedOperators[type(node.ops[0])]
            outer.subexprs.append(expr)
            self.pop_state()
            self.pop_state()
            return outer
        else:
            expr.comparator = OperatorMap[type(node.ops[0])]
            self.pop_state()
            return expr

    def visit_UnaryOp(self, node):
        if type(node.op) is Not:
            expr = self.create_expr(dast.LogicalExpr, node, {"op": dast.NotOp})
            expr.subexprs.append(self.visit(node.operand))
        else:
            expr = self.create_expr(dast.UnaryExpr, node,
                                    {"op": OperatorMap[type(node.op)]})
            expr.right = self.visit(node.operand)
        self.pop_state()
        return expr

    def visit_Subscript(self, node):
        expr = self.create_expr(dast.SubscriptExpr, node)
        if isinstance(self.current_context, Assignment) or\
           isinstance(self.current_context, Delete):
            # Assignment to an index position is an update to the container:
            self.current_context = Update(self.current_context.node)
        expr.value = self.visit(node.value)
        self.current_context = Read(expr)
        expr.index = self.visit(node.slice)
        self.pop_state()
        return expr

    def visit_Index(self, node):
        return self.visit(node.value)

    def visit_Slice(self, node):
        expr = self.create_expr(dast.SliceExpr, node)
        if node.lower is not None:
            expr.lower = self.visit(node.lower)
        if node.upper is not None:
            expr.upper = self.visit(node.upper)
        if node.step is not None:
            expr.step = self.visit(node.step)
        self.pop_state()
        return expr

    def visit_ExtSlice(self, node):
        self.warn("ExtSlice in subscript not supported.", node)
        return self.context_expr(dast.PythonExpr, node, nopush=True)

    def visit_Yield(self, node):
        # Should not get here: 'yield' statements should have been handles by
        # visit_Expr
        self.error("unexpected 'yield' expression.", node)
        return self.create_expr(dast.PythonExpr, node, nopush=True)

    def visit_YieldFrom(self, node):
        # Should not get here: 'yield from' statements should have been
        # handles by visit_Expr
        self.error("unexpected 'yield from' expression.", node)
        return self.create_expr(dast.PythonExpr, node, nopush=True)

    def visit_Lambda(self, node):
        expr = self.create_expr(dast.LambdaExpr, node)
        self.signature(node.args)
        expr.body = self.visit(node.body)
        self.pop_state()
        return expr

    def visit_Ellipsis(self, node):
        return self.create_expr(dast.EllipsisExpr, node, nopush=True)

    def generator_visit(self, node):
        if isinstance(node, SetComp):
            expr = self.create_expr(dast.SetCompExpr, node)
        elif isinstance(node, ListComp):
            expr = self.create_expr(dast.ListCompExpr, node)
        elif isinstance(node, DictComp):
            expr = self.create_expr(dast.DictCompExpr, node)
        else:
            expr = self.create_expr(dast.GeneratorExpr, node)

        self.current_context = Read(expr)
        self.enter_query()
        for g in node.generators:
            try:
                expr.unlock()
                expr.conditions.append(self.parse_domain_spec(g))
                expr.lock()
                expr.conditions.extend([self.visit(i) for i in g.ifs])
            except MalformedStatementError as e:
                self.error(e.reason, e.node)
        if isinstance(node, DictComp):
            kv = dast.KeyValue(expr)
            kv.key = self.visit(node.key)
            kv.value = self.visit(node.value)
            expr.elem = kv
        else:
            expr.elem = self.visit(node.elt)
        self.leave_query()
        self.pop_state()
        return expr

    visit_ListComp = generator_visit
    visit_GeneratorExp = generator_visit
    visit_SetComp = generator_visit
    visit_DictComp = generator_visit
    del generator_visit

    def visit_IfExp(self, node):
        expr = self.create_expr(dast.IfExpr, node)
        expr.condition = self.visit(node.test)
        expr.body = self.visit(node.body)
        expr.orbody = self.visit(node.orelse)
        self.pop_state()
        return expr

    def visit_Starred(self, node):
        expr = self.create_expr(dast.StarredExpr, node)
        expr.value = self.visit(node.value)
        self.pop_state()
        return expr

    # Helper Nodes

    def error(self, mesg, node):
        self.errcnt += 1
        if node is not None:
            printe(mesg, node.lineno, node.col_offset, self.filename)
        else:
            printe(mesg, 0, 0, self.filename)

    def warn(self, mesg, node):
        self.warncnt += 1
        if node is not None:
            printw(mesg, node.lineno, node.col_offset, self.filename)
        else:
            printw(mesg, 0, 0, self.filename)

    def debug(self, mesg, node=None):
        if node is not None:
            printd(mesg, node.lineno, node.col_offset, self.filename)
        else:
            printd(mesg, 0, 0, self.filename)

if __name__ == "__main__":
    pass



# Copyright (c) 2010-2016 Bo Lin
# Copyright (c) 2010-2016 Yanhong Annie Liu
# Copyright (c) 2010-2016 Stony Brook University
# Copyright (c) 2010-2016 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import sys
import copy
import time
import os.path
import logging
import warnings
import threading

from collections import abc, deque, namedtuple
from inspect import signature, Parameter
from functools import wraps

MAJOR_VERSION = 1
MINOR_VERSION = 0
PATCH_VERSION = 6
PRERELEASE_VERSION = ""
__version__ = "{}.{}.{}{}".format(MAJOR_VERSION, MINOR_VERSION, PATCH_VERSION,
                                   PRERELEASE_VERSION)

INCOQ_MODULE_NAME = "incoq.mars.runtime"
CONSOLE_LOG_FORMAT = \
    '[%(relativeCreated)d] %(name)s<%(processName)s>:%(levelname)s: %(message)s'
FILE_LOG_FORMAT = \
    '[%(asctime)s] %(name)s<%(processName)s>:%(levelname)s: %(message)s'

# a dict that contains the runtime configuration values:
GlobalOptions = None
# a dict containing configuration overrides set by the `config` API:
GlobalConfig = dict()
# Process id of the node process:
CurrentNode = None
# incoq.runtime.Type, only if using incoq:
IncOQBaseType = None

# Define custom levels for text output from user code using the `output`,
# `debug`, and `error` builtins, to differentiate output from user programs and
# log messages from the DistAlgo system:
logging.addLevelName(logging.INFO+1, "OUTPUT")
logging.addLevelName(logging.INFO+2, "USRERR")
logging.addLevelName(logging.DEBUG+1, "USRDBG")

log = logging.getLogger(__name__)

api_registry = dict()
builtin_registry = dict()
internal_registry = dict()

class InvalidStateException(RuntimeError): pass

def get_runtime_option(key, default=None):
    if GlobalOptions is None:
        if default is None:
            raise InvalidStateException("DistAlgo is not initialized.")
        else:
            return default

    return GlobalOptions.get(key, default)

def set_runtime_option(key, value):
    if GlobalOptions is None:
        raise InvalidStateException("DistAlgo is not initialized.")

    GlobalOptions[key] = value

def _version_as_bytes():
    """Return a 4-byte representation of the version.
    """
    prerelease = sum(ord(c) for c in PRERELEASE_VERSION) % 256
    return (((MAJOR_VERSION & 0xff) << 24) | ((MINOR_VERSION & 0xff) << 16) |
            ((PATCH_VERSION & 0xff) << 8) | prerelease).to_bytes(4, 'big')
VERSION_BYTES = _version_as_bytes()

def initialize_runtime_options(configs):
    global GlobalOptions
    if GlobalOptions is not None:
        raise InvalidStateException("DistAlgo is already initialized")

    GlobalOptions = dict(configs)

def _restore_runtime_options(params):
    global GlobalOptions, GlobalConfig
    GlobalOptions, GlobalConfig = params

def set_global_config(props):
    GlobalConfig.update(props)

def global_config():
    return GlobalConfig

def _set_node(node_id):
    global CurrentNode
    CurrentNode = node_id

def pid_of_node():
    return CurrentNode

def get_inc_module():
    if GlobalOptions is None:
        raise InvalidStateException("DistAlgo is not initialized.")
    if not hasattr(sys.modules, GlobalOptions['inc_module_name']):
        return None
    return sys.modules[GlobalOptions['inc_module_name']]

def sysinit():
    pid_format = get_runtime_option('pid_format')
    if pid_format == 'full':
        ProcessId.__str__ = ProcessId.__repr__ = ProcessId._full_form_
    elif pid_format == 'long':
        ProcessId.__str__ = ProcessId.__repr__ = ProcessId._long_form_
    else:
        # default is short
        pass
    load_modules()

def _restore_module_logging():
    assert '_da_module_cache' in GlobalOptions
    for modulename in GlobalOptions['_da_module_cache']:
        consolefmt, filefmt = GlobalOptions['_da_module_cache'][modulename]
        setup_logging_for_module(modulename, consolefmt, filefmt)

def setup_logging_for_module(modulename,
                             consolefmt=CONSOLE_LOG_FORMAT,
                             filefmt=FILE_LOG_FORMAT):
    """Configures package level logger.

    """
    if GlobalOptions is None:
        # We're not yet initialized, which will happen when using `import_da`
        # under spawning semantics. This is fine, as logging will be setup after
        # `OSProcessContainer.run` gets called. We can safely ignore this call:
        return
    if '_da_module_cache' not in GlobalOptions:
        GlobalOptions['_da_module_cache'] = dict()
    GlobalOptions['_da_module_cache'][modulename] = (consolefmt, filefmt)
    rootlog = logging.getLogger(modulename)
    rootlog.handlers = []       # Clear all handlers

    if not GlobalOptions['no_log']:
        rootlog.propagate = False
        rootlog.setLevel(logging.DEBUG)
        consoleformatter = logging.Formatter(consolefmt)
        consolelvl = logging._nameToLevel[GlobalOptions['logconsolelevel'].upper()]
        ch = logging.StreamHandler()
        ch.setFormatter(consoleformatter)
        ch.setLevel(consolelvl)
        rootlog._consolelvl = consolelvl
        rootlog.addHandler(ch)

        if GlobalOptions['logfile']:
            filelvl = logging._nameToLevel[GlobalOptions['logfilelevel'].upper()]
            logfilename = GlobalOptions['logfilename']
            if logfilename is None:
                if GlobalOptions['file'] is not None:
                    logfilename = \
                             (os.path.basename(GlobalOptions['file']) + ".log")
                else:
                    logfilename = GlobalOptions['module'][0] + '.log'
            fh = logging.FileHandler(logfilename)
            formatter = logging.Formatter(filefmt)
            fh.setFormatter(formatter)
            fh.setLevel(filelvl)
            rootlog._filelvl = filelvl
            rootlog.addHandler(fh)

        if GlobalOptions['logdir'] is not None:
            os.makedirs(GlobalOptions['logdir'], exist_ok=True)
            rootlog._logdir = GlobalOptions['logdir']
        else:
            rootlog._logdir = None
    else:
        rootlog.addHandler(logging.NullHandler())

def load_modules():
    import importlib
    global IncOQBaseType
    if not GlobalOptions['load_inc_module']:
        return
    main = sys.modules[GlobalOptions['this_module_name']]
    inc = importlib.import_module(GlobalOptions['inc_module_name'])
    if inc.JbStyle:
        IncOQBaseType = importlib.import_module(INCOQ_MODULE_NAME) \
                        .IncOQType
    if GlobalOptions['control_module_name'] is not None:
        ctrl = importlib.import_module(GlobalOptions['control_module_name'])
        main.IncModule = ModuleIntrument(ctrl, inc)
    else:
        main.IncModule = inc

####################
# Process ID
####################
ILLEGAL_NAME_CHARS = set('@#:')

def check_name(name):
    return not (set(name) & ILLEGAL_NAME_CHARS)

def name_split_host(name):
    """Splits `name` into 'processname', 'hostname', and 'port' components."""
    comps = name.split('@')
    if len(comps) == 2:
        name, suffix = comps
        suffix = suffix.split(':')
        if len(suffix) > 2:
            return (None, None, None)
        elif len(suffix) == 2:
            host, port = suffix
            try:
                return (name, host, int(port))
            except ValueError:
                return (None, None, None)
        elif len(suffix) == 1:
            return (name, suffix[0], None)
        else:
            return (name, None, None)
    elif len(comps) == 1:
        return (comps[0], None, None)
    else:
        return (None, None, None)

def name_split_node(name):
    """Splits `name` into 'processname', 'nodename' components."""
    assert '#' not in name
    comps = name.split('#')
    if len(comps) == 2:
        return tuple(comps)
    elif len(comps) == 1:
        return (comps[0], comps[0])
    else:
        return (None, None)

class ProcessId(namedtuple("_ProcessId",
                           'uid, seqno, pcls, \
                           name, nodename, hostname, transports')):
    """An instance of `ProcessId` uniquely identifies a DistAlgo process instance.

    A `ProcessId` instance should contain all necessary information for any
    DistAlgo process in a distributed system to send messages to that process.
    This includes the network addresses of all ports the process listens on.

    There is a total ordering over the set of all `ProcessId`s. `ProcessId`s
    referring to the same process will always compare equal.

    From the point of view of DistAlgo programs, `ProcessId` instances are
    opaque objects -- no assumptions should be made about the internal structure
    of `ProcessId` instances.

    """
    __slots__ = ()
    _pid_counter = 0
    _lock = threading.Lock()
    _named = dict()
    _callbacks = dict()

    def __new__(cls, uid, seqno, pcls, name, nodename, hostname, transports):
        obj = super().__new__(cls, uid, seqno, pcls, name, nodename, hostname,
                              transports)
        if len(name) > 0:
            fullname = (name, nodename)
            with ProcessId._lock:
                entry = ProcessId._named.get(fullname, None)
                callbacks = ProcessId._callbacks.get(fullname, None)
                if isinstance(entry, ProcessId):
                    if obj < entry:
                        # cached id is more recent than the new one, so use the
                        # cached entry:
                        obj = entry
                    elif obj.uid != entry.uid:
                        log.warning("Process name '%s#%s' reassigned from %s "
                                    "to %s.", name, nodename,
                                    ProcessId._full_form_(entry),
                                    ProcessId._full_form_(obj))
                if entry != obj:
                    ProcessId._named[fullname] = obj
                if callbacks is not None:
                    del ProcessId._callbacks[fullname]
            if type(callbacks) is list:
                for callback in callbacks:
                    assert callable(callback)
                    callback(obj)
        return obj

    @staticmethod
    def lookup(name):
        return ProcessId._named.get(name, None)

    @staticmethod
    def lookup_or_register_callback(name, callback):
        with ProcessId._lock:
            if name not in ProcessId._named:
                if name not in ProcessId._callbacks:
                    ProcessId._callbacks[name] = [callback]
                else:
                    ProcessId._callbacks[name].append(callback)
                return None
            else:
                return ProcessId._named[name]

    @staticmethod
    def all_named_ids():
        with ProcessId._lock:
            return list(ProcessId._named.values())

    @staticmethod
    def drop_entry(nid):
        with ProcessId._lock:
            if nid.name in ProcessId._named:
                del ProcessId._named[nid.name]

    @staticmethod
    def gen_uid(hostname, pid):
        """Generate a globally unique 96-bit id.

        """
        # 54 bits of timestamp:
        tstamp = int(time.time() * 1000) & 0x3fffffffffffff
        # 16 bits of hostname hash
        hh = int(hash(hostname)) & 0xffff
        # 16 bits of os pid
        pid %= 0xffff
        # 10 bit global counter
        with ProcessId._lock:
            cnt = ProcessId._pid_counter = (ProcessId._pid_counter + 1) % 1024
        return (tstamp << 42) | (hh << 26) | (pid << 10) | cnt

    @classmethod
    def _create(idcls, pcls, transports, name=""):
        """Creates a new `ProcessId` instance.

        """
        hostname = get_runtime_option('hostname')
        nodename = get_runtime_option('nodename')
        uid = ProcessId.gen_uid(hostname,
                                pid=threading.current_thread().ident)
        return idcls(uid=uid, seqno=1, pcls=pcls,
                     name=name, nodename=nodename,
                     hostname=hostname, transports=transports)

    def _short_form_(self):
        """Constructs a short string representation of this pid.

        This form is more suitable for use in output strings.

        """
        if len(self.nodename) > 0 and self.nodename != self.name:
            if len(self.name) > 0:
                return "<{0.pcls.__name__}:{0.name}#{0.nodename}>".format(self)
            else:
                # otherwise, we use `uid` truncated to the last 5 hex digits:
                return "<{0.pcls.__name__}:{1:05x}#{0.nodename}>".format(
                    self, self.uid & 0xfffff)
        else:
            if len(self.name) > 0:
                return "<{0.pcls.__name__}:{0.name}>".format(self)
            else:
                return "<{0.pcls.__name__}:{1:05x}>".format(self, self.uid & 0xfffff)

    def _long_form_(self):
        """Constructs a short string representation of this pid.

        This form is more suitable for use in output strings.

        """
        if len(self.nodename) > 0 and self.nodename != self.name:
            if len(self.name) > 0:
                return "<{0.pcls.__name__}:{0.name}#{0.nodename}>".format(self)
            else:
                # otherwise, we use `uid` truncated to the last 5 hex digits:
                return "<{0.pcls.__name__}:{1:x}#{0.nodename}>".format(self, self.uid)
        else:
            if len(self.name) > 0:
                return "<{0.pcls.__name__}:{0.name}>".format(self)
            else:
                return "<{0.pcls.__name__}:{1:x}>".format(self, self.uid)

    def _full_form_(self):
        """Constructs a full string representation of this pid.

        This form may be more useful in debugging.

        """
        fmt = "ProcessId(uid={0.uid:x}, seqno={0.seqno}, " \
              "pcls={0.pcls.__name__}, " \
              "name='{0.name}', nodename='{0.nodename}', " \
              "hostname='{0.hostname}', transports={0.transports})"
        return fmt.format(self)

    __str__ = __repr__ = _short_form_

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

####################

warnings.simplefilter("default", DeprecationWarning)
def deprecated(func):
    """Declare 'func' as deprecated.

    This is a decorator which can be used to mark functions as deprecated. It
    will result in a warning being emmitted when the function is used.

    """
    def newFunc(*args, **kwargs):
        warnings.warn("Call to deprecated function %s." % func.__name__,
                      category=DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)
    newFunc.__name__ = func.__name__
    newFunc.__doc__ = func.__doc__
    newFunc.__dict__.update(func.__dict__)
    return newFunc

def api(func):
    """Declare 'func' as DistPy API.

    This wraps the function to perform basic type checking for type-annotated
    parameters and return value.

    """
    global api_registry
    funame = func.__name__
    if api_registry.get(funame) is not None:
        return api_registry[funame]

    sig = signature(func)

    @wraps(func)
    def _func_impl(*args, **kwargs):
        try:
            binding = sig.bind(*args, **kwargs)
        except TypeError as e:
            log.error(str(e))
            return None
        for argname in binding.arguments:
            atype = sig.parameters[argname].annotation
            if (atype is not Parameter.empty and
                    not isinstance(binding.arguments[argname], atype)):
                log.error(
                    "'%s' called with wrong type argument: "
                     "%s, expected %s, got %s.",
                    funame, argname, str(atype),
                    str(binding.arguments[argname].__class__))
                return None
        result = func(*args, **kwargs)
        if (sig.return_annotation is not Parameter.empty and
                not isinstance(result, sig.return_annotation)):
            log.warning(
                "Possible bug: API function '%s' return value type mismatch: "
                "declared %s, returned %s.",
                funame, sig.return_annotation, result.__class__)
        return result

    _func_impl.__name__ = func.__name__
    _func_impl.__doc__ = func.__doc__
    _func_impl.__dict__.update(func.__dict__)
    api_registry[funame] = _func_impl
    return _func_impl

def builtin(func):
    """Declare `func` as DistAlgo builtin.

    Builtins are instance methods of da.DistProcess, and must be called with
    the process instance as first argument (self).

    """
    funame = func.__name__
    if builtin_registry.get(funame) is not None:
        return builtin_registry[funame]
    else:
        builtin_registry[funame] = func
        return func

def internal(func):
    """Declare `func` as `DistProcess` internal implementation.

    This gives the compiler a hint to prevent user code from unintentionally
    overriding an internal function.

    """
    funame = func.__name__
    internal_registry[funame] = func
    return func

class Namespace(object):
    pass

class Null(object):
    def __init__(self, *args, **kwargs): pass
    def __call__(self, *args, **kwargs): return self
    def __getattribute__(self, attr): return self
    def __setattr__(self, attr, value): pass
    def __delattr__(self, attr): pass

class BufferIOWrapper:
    def __init__(self, barray):
        self.buffer = barray
        self.total_bytes = len(barray)
        self.fptr = 0

    def write(self, data):
        end = self.fptr + len(data)
        if end > self.total_bytes:
            raise IOError("buffer full.")
        self.buffer[self.fptr:end] = data
        self.fptr = end

class QueueEmpty(Exception): pass
class WaitableQueue:
    """This class implements a fast waitable queue based on a `deque`.

    The `Queue` class from the standard library `queue` module has a lot of
    unneccesary overhead due to synchronization on every `get` and `put`
    operation. We can avoid these synchronization overheads by piggy-backing off
    the CPython GIL when the queue is non-empty.

    """
    def __init__(self, iterable=[], maxlen=None):
        self._q = deque(iterable, maxlen)
        self._condition = threading.Condition()
        self._num_waiting = 0

    def append(self, item):
        self._q.append(item)
        if self._num_waiting > 0:
            with self._condition:
                self._condition.notify_all()

    def pop(self, block=True, timeout=None):
        # Opportunistically try to get the next item off the queue:
        try:
            return self._q.popleft()
        except IndexError:
            pass
        # The queue was empty, if we don't need to block then we're done:
        if not block or timeout == 0:
            raise QueueEmpty()
        # Otherwise, we have to acquire the condition object and block:
        try:
            with self._condition:
                self._num_waiting += 1
                self._condition.wait(timeout)
                self._num_waiting -= 1
            return self._q.popleft()
        except IndexError:
            # If the queue is still empty at this point, it means that the new
            # event was picked up by another thread, so it's ok for us to
            # return:
            raise QueueEmpty()
        # Other exceptions will be propagated

    def __len__(self):
        return self._q.__len__()


class Node(object):
    __slots__ = ['prev', 'next', 'me']
    def __init__(self, prev, me):
        self.prev = prev
        self.me = me
        self.next = None
    def __str__(self):
        return str(self.me)
    def __repr__(self):
        return self.me.__repr__()

class LRU:
    """
    Implementation of a length-limited O(1) LRU queue.
    Built for and used by PyPE:
    http://pype.sourceforge.net
    Copyright 2003 Josiah Carlson.
    """
    def __init__(self, count, pairs=[]):
        self.count = max(count, 1)
        self.d = {}
        self.first = None
        self.last = None
        for key, value in pairs:
            self[key] = value
    def __contains__(self, obj):
        return obj in self.d
    def __getitem__(self, obj):
        a = self.d[obj].me
        self[a[0]] = a[1]
        return a[1]
    def __setitem__(self, obj, val):
        if obj in self.d:
            del self[obj]
        nobj = Node(self.last, (obj, val))
        if self.first is None:
            self.first = nobj
        if self.last:
            self.last.next = nobj
        self.last = nobj
        self.d[obj] = nobj
        if len(self.d) > self.count:
            if self.first == self.last:
                self.first = None
                self.last = None
                return
            a = self.first
            a.next.prev = None
            self.first = a.next
            a.next = None
            del self.d[a.me[0]]
            del a
    def __delitem__(self, obj):
        nobj = self.d[obj]
        if nobj.prev:
            nobj.prev.next = nobj.next
        else:
            self.first = nobj.next
        if nobj.next:
            nobj.next.prev = nobj.prev
        else:
            self.last = nobj.prev
        del self.d[obj]
    def __iter__(self):
        cur = self.first
        while cur != None:
            cur2 = cur.next
            yield cur.me[1]
            cur = cur2
    def __str__(self):
        return str(self.d)
    def __repr__(self):
        return self.d.__repr__()
    def iteritems(self):
        cur = self.first
        while cur != None:
            cur2 = cur.next
            yield cur.me
            cur = cur2
    def iterkeys(self):
        return iter(self.d)
    def itervalues(self):
        for i,j in self.iteritems():
            yield j
    def keys(self):
        return self.d.keys()
    def get(self, k, d=None):
        v = self.d.get(k)
        if v is None: return None
        a = v.me
        self[a[0]] = a[1]
        return a[1]


class IntrumentationError(Exception): pass
class FunctionInstrument(object):
    def __init__(self, control_func, subject_func):
        super().__setattr__('_control', control_func)
        super().__setattr__('_subject', subject_func)

    def __call__(self, *args, **kwargs):
        ctrl_result = self._control(*args, **kwargs)
        subj_result = self._subject(*args, **kwargs)
        if ctrl_result != subj_result:
            raise IntrumentationError("Result mismatch for %s: "
                                      "control returned %s; "
                                      "subject returned %s." %
                                      (self._control.__name__,
                                       str(ctrl_result),
                                       str(subj_result)))
        return subj_result

    def __setattr__(self, attr, value):
        super().__setattr__(attr, value)
        setattr(self._control, attr, value)
        setattr(self._subject, attr, value)

    def __delattr__(self, attr):
        super().__delattr__(attr)
        delattr(self._control, attr)
        delattr(self._subject, attr)

class ModuleIntrument(object):
    def __init__(self, control, subject):
        super().__setattr__('_control', control)
        super().__setattr__('_subject', subject)

    def __getattribute__(self, attr):
        ctrl_attr = getattr(super().__getattribute__('_control'), attr)
        subj_attr = getattr(super().__getattribute__('_subject'), attr)
        if type(ctrl_attr) is not type(subj_attr):
            raise IntrumentationError("Attribute mismatch for %s:"
                                      "control is type %s; "
                                      "subject is type %s." %
                                      (attr, str(type(ctrl_attr)),
                                       str(type(subj_attr))))
        if hasattr(ctrl_attr, '__call__') and \
           (ctrl_attr.__name__.startswith("Query_") or
            ctrl_attr.__name__ == "init"):
                return FunctionInstrument(ctrl_attr, subj_attr)
        else:
            return subj_attr

    def __setattr__(self, attr, value):
        super().__setattr__(attr, value)
        setattr(self._control, attr, value)
        setattr(self._subject, attr, value)

    def __delattr__(self, attr):
        super().__delattr__(attr)
        delattr(self._control, attr)
        delattr(self._subject, attr)

class frozendict(dict):
    """Hashable immutable dict implementation

    Copied from http://code.activestate.com/recipes/414283/

    """
    def _blocked_attribute(obj):
        raise AttributeError("A frozendict cannot be modified.")
    _blocked_attribute = property(_blocked_attribute)

    __delitem__ = __setitem__ = clear = _blocked_attribute
    pop = popitem = setdefault = update = _blocked_attribute

    def __new__(cls, *args, **kws):
        new = dict.__new__(cls)
        dict.__init__(new, *args, **kws)
        return new

    def __init__(self, *args, **kws):
        pass

    def __hash__(self):
        try:
            return self._cached_hash
        except AttributeError:
            h = self._cached_hash = hash(tuple(sorted(self.items())))
            return h

    def __repr__(self):
        return "frozendict(%s)" % dict.__repr__(self)

BuiltinImmutables = [int, float, complex, tuple, str, bytes, frozenset]

def freeze(obj):
    """Return a hashable version of `obj`.

    Contents of `obj` may be copied if necessary.

    """
    if any(isinstance(obj, itype) for itype in BuiltinImmutables):
        return obj

    if IncOQBaseType is not None:
        if isinstance(obj, IncOQBaseType):
            return copy.deepcopy(obj)

    if isinstance(obj, abc.MutableSequence):
        if isinstance(obj, abc.ByteString):
            # bytearray -> bytes
            return bytes(obj)
        else:
            # list -> tuple
            return tuple(freeze(elem) for elem in obj)
    elif isinstance(obj, abc.MutableSet):
        # set -> frozenset
        return frozenset(freeze(elem) for elem in obj)
    elif isinstance(obj, abc.MutableMapping):
        # dict -> frozendict
        return frozendict((freeze(k), freeze(v)) for k, v in obj.items())
    elif isinstance(obj, abc.Sequence):
        #NOTE: This part is fragile. For immutable sequence objects, we still
        # have to recursively freeze its elements, which means we have to create
        # a new instance of the same type. Here we just assume the class
        # `__init__` method takes a 'iterable' as argument. Otherwise,
        # everything falls apart.
        return type(obj)(freeze(e) for e in obj)
    else:
        # everything else just assume hashable & immutable, hahaha:
        return obj

if __name__ == "__main__":
    @api
    def testapi(a : int, b : list) -> dict:
        print (a, b)
        return []

    testapi(1, [2])
    testapi(1, {})
    print(api_registry)

    @deprecated
    def testdepre():
        print("deprecated function")

    t = ('a', 1)
    print("Freeze " + str(t) + "->" + str(freeze(t)))
    testdepre()



from .common import freeze

class PatternElement:
    """Tree structure representing a message pattern.
    """

    def __init__(self, value):
        self.value = value

    def filter(self, iterable, order=None, **context):
        if isinstance(iterable, set):
            return self._filter_set(iterable, order, **context)
        else:
            return self._filter_list(iterable, order, **context)

    def _filter_list(self, iterable, order=None, **context):
        for elt in iterable:
            bindings = dict()
            if self.match(elt, bindings=bindings, **context):
                if len(bindings) > 0 and order is not None:
                    yield tuple(bindings[name] for name in order)
                else:
                    yield True

    def _filter_set(self, iterable, order=None, **context):
        seen = set()
        for elt in iterable:
            bindings = dict()
            if self.match(elt, bindings=bindings, **context):
                if len(bindings) > 0 and order is not None:
                    res = tuple(bindings[name] for name in order)
                    if res not in seen:
                        seen.add(res)
                        yield res
                else:
                    if True not in seen:
                        seen.add(True)
                        yield True

    def match(self, message, bindings=None,
              ignore_bound_vars=False, **context):
        return False

    def match_iter(self, iterable, **context):
        for elt in iterable:
            bindings = dict()
            if self.match(elt, bindings=bindings, **context):
                return True
        return False

    def __str__(self):
        raise "<PatternElement " + str(self.value) + ">"

    def __repr__(self):
        return str(self)

class ConstantPattern(PatternElement):
    def match(self, message, bindings=None,
              ignore_bound_vars=False, **context):
        return self.value == message

    def __str__(self):
        return "=" + repr(self.value)

class SelfPattern(ConstantPattern):
    def __init__(self):
        super().__init__(None)

    def match(self, message, bindings=None,
              ignore_bound_vars=False, SELF_ID=None, **context):
        return message == SELF_ID

    def __str__(self):
        return "=SELF"

class BoundPattern(PatternElement):
    def match(self, message, bindings=None,
              ignore_bound_vars=False, **context):
        if ignore_bound_vars:
            return True
        try:
            return context[self.value] == message
        except KeyError:
            return False

    def __str__(self):
        return "=" + self.value

class FreePattern(PatternElement):
    def match(self, message, bindings=None,
              ignore_bound_vars=False, **context):
        if bindings is None:
            bindings = dict()
        if self.value is not None:
            try:
                if bindings[self.value] != message:
                    return False
            except KeyError:
                bindings[self.value] = message
        return True

    def __str__(self):
        if self.value is None:
            return "_"
        else:
            return self.value

class TuplePattern(PatternElement):
    def match(self, message, bindings=None,
              ignore_bound_vars=False, **context):
        if bindings is None:
            bindings = dict()
        if (type(message) is not tuple or
            len(message) != len(self.value)):
            return False
        # Fall through
        for pat, mem in zip(self.value, message):
            if not pat.match(mem, bindings, ignore_bound_vars, **context):
                return False
        return True

    def __str__(self):
        return "(" + ",".join([str(p) for p in self.value]) + ")"

class ListPattern(PatternElement):
    def match(self, message, bindings=None,
              ignore_bound_vars=False, **context):
        if bindings is None:
            bindings = dict()
        if (type(message) is not list or
            len(message) != len(self.value)):
            return False
        # Fall through
        for pat, mem in zip(self.value, message):
            if not pat.match(mem, bindings, ignore_bound_vars, **context):
                return False
        return True

    def __str__(self):
        return "[" + ",".join([str(p) for p in self.value]) + "]"


class Event:
    """ Describes a single event.

    Instances of Event are created by the backend thread and passed to the
    front end.
    """
    def __init__(self, envelope, message):
        (self.timestamp, self.destination, self.source) = envelope
        self.message = message

    def to_tuple(self):
        """Generates a tuple representation for this event."""
        return (type(self),
                (self.timestamp, self.destination, self.source),
                freeze(self.message))

    def __str__(self):
        buf = ["<", type(self).__name__,
               " time:", str(self.timestamp),
               " to:", repr(self.destination),
               " from:", repr(self.source),
               " msg:", repr(self.message),
               ">"]
        return "".join(buf)

    def __repr__(self):
        return str(self)

class ReceivedEvent(Event): pass
class SentEvent(Event): pass


class EventPattern:
    """ Describes an event "pattern" that can be used to match against Event
    instances.
    """

    def __init__(self, eventtype, name, pattern,
                 sources=None, destinations=None, timestamps=None,
                 record_history=False, handlers=[]):
        self.eventtype = eventtype
        self.name = name
        self.pattern = pattern
        self.sources = sources
        self.destinations = destinations
        self.timestamps = timestamps
        self.record_history = record_history
        self.handlers = handlers

    def filter(self, iterable, order=None, **context):
        for elt in iterable:
            bindings = dict()
            if self.match(elt, bindings=bindings, **context):
                if len(bindings) > 0 and order is not None:
                    yield tuple(bindings[name] for name in order)
                else:
                    yield True

    def match(self, event, bindings=None,
              ignore_bound_vars=False, **context):
        if isinstance(event, tuple):
            event = event[0](*event[1:])
        if type(event) is not self.eventtype:
            return False
        if bindings is None:
            bindings = dict()

        if self.sources is not None:
            for pat in self.sources:
                if pat.match(event.source, bindings,
                             ignore_bound_vars, **context):
                    break
            else:
                return False

        if self.destinations is not None:
            for pat in self.destinations:
                if pat.match(event.destination, bindings,
                             ignore_bound_vars, **context):
                    break
            else:
                return False

        if self.timestamps is not None:
            for pat in self.timestamps:
                if pat.match(event.timestamp, bindings,
                             ignore_bound_vars, **context):
                    break
            else:
                return False

        if (self.pattern is not None and
                not self.pattern.match(event.message, bindings,
                                       ignore_bound_vars, **context)):
            return False
        else:
            return True

    def __str__(self):
        buf = ["<", self.eventtype.__name__,
               " name:", self.name,
               " msg:", str(self.pattern)]
        if self.sources is not None:
            buf.extend([" from:{",
                        ",".join([str(s) for s in self.sources]),
                        "}"])
        if self.destinations is not None:
            buf.extend([" to:{",
                        ",".join([str(s) for s in self.destinations]),
                        "}"])
        if self.timestamps is not None:
            buf.extend([" time:{",
                        ",".join([str(s) for s in self.timestamps]),
                        "}"])
        buf.extend([">"])
        return "".join(buf)

    def __repr__(self):
        return str(self)


if __name__ == "__main__":
    # pattern = PatternElement(
    #     TupleVar,
    #     [PatternElement(ConstantVar, "xxx"),
    #      PatternElement(BoundVar, "foo1"),
    #      PatternElement(
    #          ListVar,
    #          [PatternElement(FreeVar, "bar1"),
    #           PatternElement(ConstantVar, 5)]
    #          )
    #  ])
    eventpat = EventPattern(ReceivedEvent,
                            'ReceivedEvent_1',
                            TuplePattern([ConstantPattern('Request'),
                                          FreePattern('reqts')]),
                            sources=[FreePattern('source')],
                            destinations=None,
                            timestamps=None,
                            record_history=False, handlers=[])
    # evtpat = EventPattern(ReceiveEvent,
    #                       "abc",
    #                       pattern,
    #                       sources={PatternElement(FreeVar, "src1")},
    #                       destinations={PatternElement(FreeVar, "self")},
    #                       timestamps=None
    #                       )
    message1 = ("Request", 0)
    evt = ReceivedEvent(message1, 0, 1, 2)
    b = dict()
    res = eventpat.match(evt, bindings=b, **{"foo1": 2})
    print(eventpat, evt)



# -*- generated by 1.0.6 -*-
import da
PatternExpr_215 = da.pat.TuplePattern([da.pat.ConstantPattern('respond'), da.pat.BoundPattern('_BoundPattern218_'), da.pat.FreePattern(None)])
PatternExpr_222 = da.pat.FreePattern('a')
PatternExpr_250 = da.pat.TuplePattern([da.pat.ConstantPattern('respond'), da.pat.BoundPattern('_BoundPattern253_'), da.pat.TuplePattern([da.pat.FreePattern('n2'), da.pat.FreePattern('v')])])
PatternExpr_281 = da.pat.TuplePattern([da.pat.ConstantPattern('respond'), da.pat.BoundPattern('_BoundPattern284_'), da.pat.TuplePattern([da.pat.FreePattern('n2'), da.pat.FreePattern(None)])])
PatternExpr_317 = da.pat.TuplePattern([da.pat.ConstantPattern('respond'), da.pat.BoundPattern('_BoundPattern320_'), da.pat.FreePattern(None)])
PatternExpr_324 = da.pat.FreePattern('a')
PatternExpr_365 = da.pat.TuplePattern([da.pat.ConstantPattern('done')])
PatternExpr_370 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('done')])])
PatternExpr_410 = da.pat.TuplePattern([da.pat.ConstantPattern('prepare'), da.pat.FreePattern('n')])
PatternExpr_417 = da.pat.FreePattern('p')
PatternExpr_423 = da.pat.TuplePattern([da.pat.ConstantPattern('respond'), da.pat.FreePattern('n2'), da.pat.FreePattern(None)])
PatternExpr_452 = da.pat.TuplePattern([da.pat.ConstantPattern('accepted'), da.pat.FreePattern('n'), da.pat.FreePattern('v')])
PatternExpr_478 = da.pat.TuplePattern([da.pat.ConstantPattern('accepted'), da.pat.FreePattern('n'), da.pat.FreePattern(None)])
PatternExpr_513 = da.pat.TuplePattern([da.pat.ConstantPattern('accept'), da.pat.FreePattern('n'), da.pat.FreePattern('v')])
PatternExpr_526 = da.pat.TuplePattern([da.pat.ConstantPattern('respond'), da.pat.FreePattern('n2'), da.pat.FreePattern(None)])
PatternExpr_562 = da.pat.TuplePattern([da.pat.ConstantPattern('done')])
PatternExpr_567 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('done')])])
PatternExpr_610 = da.pat.TuplePattern([da.pat.ConstantPattern('accepted'), da.pat.FreePattern('n'), da.pat.FreePattern('v')])
PatternExpr_636 = da.pat.TuplePattern([da.pat.ConstantPattern('accepted'), da.pat.BoundPattern('_BoundPattern639_'), da.pat.BoundPattern('_BoundPattern640_')])
PatternExpr_643 = da.pat.FreePattern('a')
PatternExpr_788 = da.pat.TuplePattern([da.pat.ConstantPattern('learned')])
PatternExpr_793 = da.pat.BoundPattern('_BoundPattern794_')
PatternExpr_795 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern801_')]), da.pat.TuplePattern([da.pat.ConstantPattern('learned')])])
_config_object = {}
import sys
from random import randint
TIMEOUT = 1

class Proposer(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._ProposerReceivedEvent_0 = []
        self._ProposerReceivedEvent_1 = []
        self._ProposerReceivedEvent_2 = []
        self._ProposerReceivedEvent_3 = []
        self._ProposerReceivedEvent_4 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_0', PatternExpr_215, sources=[PatternExpr_222], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_1', PatternExpr_250, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_2', PatternExpr_281, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_3', PatternExpr_317, sources=[PatternExpr_324], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_ProposerReceivedEvent_4', PatternExpr_365, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, acceptors, **rest_814):
        super().setup(acceptors=acceptors, **rest_814)
        self._state.acceptors = acceptors
        self._state.n = None
        self._state.majority = self._state.acceptors

    def run(self):
        while (not PatternExpr_370.match_iter(self._ProposerReceivedEvent_4, SELF_ID=self._id)):
            self.to_consent()
        self.output('terminating')

    def to_consent(self):
        self._state.n = ((0, self._id) if (self._state.n == None) else ((self._state.n[0] + 1), self._id))
        self.send(('prepare', self._state.n), to=self._state.majority)
        super()._label('_st_label_210', block=False)
        _st_label_210 = 0
        self._timer_start()
        while (_st_label_210 == 0):
            _st_label_210 += 1
            if (len({a for (_, (_, _, a), (_ConstantPattern233_, _BoundPattern235_, _)) in self._ProposerReceivedEvent_0 if (_ConstantPattern233_ == 'respond') if (_BoundPattern235_ == self._state.n)}) > (len(self._state.acceptors) / 2)):
                v = self.anyof(({v for (_, _, (_ConstantPattern269_, _BoundPattern271_, (n2, v))) in self._ProposerReceivedEvent_1 if (_ConstantPattern269_ == 'respond') if (_BoundPattern271_ == self._state.n) if (n2 == max({n2 for (_, _, (_ConstantPattern298_, _BoundPattern300_, (n2, _))) in self._ProposerReceivedEvent_2 if (_ConstantPattern298_ == 'respond') if (_BoundPattern300_ == self._state.n)}))} or {randint(1, 100)}))
                responded = {a for (_, (_, _, a), (_ConstantPattern335_, _BoundPattern337_, _)) in self._ProposerReceivedEvent_3 if (_ConstantPattern335_ == 'respond') if (_BoundPattern337_ == self._state.n)}
                self.send(('accept', self._state.n, v), to=responded)
                self.debug('### chose', self._state.n, v)
                _st_label_210 += 1
            elif self._timer_expired:
                self.output('failed proposal number', self._state.n)
                _st_label_210 += 1
            else:
                super()._label('_st_label_210', block=True, timeout=TIMEOUT)
                _st_label_210 -= 1

    def anyof(self, s):
        return (next(iter(s)) if s else None)

class Acceptor(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._AcceptorSentEvent_1 = []
        self._AcceptorSentEvent_2 = []
        self._AcceptorSentEvent_3 = []
        self._AcceptorSentEvent_5 = []
        self._AcceptorReceivedEvent_6 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_0', PatternExpr_410, sources=[PatternExpr_417], destinations=None, timestamps=None, record_history=None, handlers=[self._Acceptor_handler_409]), da.pat.EventPattern(da.pat.SentEvent, '_AcceptorSentEvent_1', PatternExpr_423, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.SentEvent, '_AcceptorSentEvent_2', PatternExpr_452, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.SentEvent, '_AcceptorSentEvent_3', PatternExpr_478, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_4', PatternExpr_513, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._Acceptor_handler_512]), da.pat.EventPattern(da.pat.SentEvent, '_AcceptorSentEvent_5', PatternExpr_526, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_AcceptorReceivedEvent_6', PatternExpr_562, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, learners, **rest_814):
        super().setup(learners=learners, **rest_814)
        self._state.learners = learners
        pass

    def run(self):
        super()._label('_st_label_559', block=False)
        _st_label_559 = 0
        while (_st_label_559 == 0):
            _st_label_559 += 1
            if PatternExpr_567.match_iter(self._AcceptorReceivedEvent_6, SELF_ID=self._id):
                _st_label_559 += 1
            else:
                super()._label('_st_label_559', block=True)
                _st_label_559 -= 1
        self.output('terminating')

    def anyof(self, s):
        "return any element of set s if s is not empty or 'None' otherwise"
        return (next(iter(s)) if s else None)

    def _Acceptor_handler_409(self, n, p):
        n2 = None

        def UniversalOpExpr_421():
            nonlocal n2
            for (_, _, (_ConstantPattern439_, n2, _)) in self._AcceptorSentEvent_1:
                if (_ConstantPattern439_ == 'respond'):
                    if (not (n > n2)):
                        return False
            return True
        if UniversalOpExpr_421():
            maxprop = self.anyof({(n, v) for (_, _, (_ConstantPattern468_, n, v)) in self._AcceptorSentEvent_2 if (_ConstantPattern468_ == 'accepted') if (n == max({n for (_, _, (_ConstantPattern493_, n, _)) in self._AcceptorSentEvent_3 if (_ConstantPattern493_ == 'accepted')}))})
            self.send(('respond', n, maxprop), to=p)
    _Acceptor_handler_409._labels = None
    _Acceptor_handler_409._notlabels = None

    def _Acceptor_handler_512(self, n, v):
        n2 = None

        def ExistentialOpExpr_524():
            nonlocal n2
            for (_, _, (_ConstantPattern542_, n2, _)) in self._AcceptorSentEvent_5:
                if (_ConstantPattern542_ == 'respond'):
                    if (n2 > n):
                        return True
            return False
        if (not ExistentialOpExpr_524()):
            self.send(('accepted', n, v), to=self._state.learners)
    _Acceptor_handler_512._labels = None
    _Acceptor_handler_512._notlabels = None

class Learner(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._LearnerReceivedEvent_0 = []
        self._LearnerReceivedEvent_1 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_LearnerReceivedEvent_0', PatternExpr_610, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_LearnerReceivedEvent_1', PatternExpr_636, sources=[PatternExpr_643], destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, acceptors, **rest_814):
        super().setup(acceptors=acceptors, **rest_814)
        self._state.acceptors = acceptors
        pass

    def run(self):
        self.learn()
        self.output('terminating')
        self.send(('learned',), to=self.nodeof(self._id))

    def learn(self):
        super()._label('_st_label_607', block=False)
        a = n = v = None

        def ExistentialOpExpr_608():
            nonlocal a, n, v
            for (_, _, (_ConstantPattern627_, n, v)) in self._LearnerReceivedEvent_0:
                if (_ConstantPattern627_ == 'accepted'):
                    if (len({a for (_, (_, _, a), (_ConstantPattern654_, _BoundPattern656_, _BoundPattern657_)) in self._LearnerReceivedEvent_1 if (_ConstantPattern654_ == 'accepted') if (_BoundPattern656_ == n) if (_BoundPattern657_ == v)}) > (len(self._state.acceptors) / 2)):
                        return True
            return False
        _st_label_607 = 0
        self._timer_start()
        while (_st_label_607 == 0):
            _st_label_607 += 1
            if ExistentialOpExpr_608():
                self.output('learned', n, v)
                _st_label_607 += 1
            elif self._timer_expired:
                self.output('failed learning anything')
                _st_label_607 += 1
            else:
                super()._label('_st_label_607', block=True, timeout=(TIMEOUT * 10))
                _st_label_607 -= 1

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._Node_ReceivedEvent_0 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_Node_ReceivedEvent_0', PatternExpr_788, sources=[PatternExpr_793], destinations=None, timestamps=None, record_history=True, handlers=[])])

    def run(self):
        nacceptors = (int(sys.argv[1]) if (len(sys.argv) > 1) else 3)
        nproposers = (int(sys.argv[2]) if (len(sys.argv) > 2) else 5)
        nlearners = (int(sys.argv[3]) if (len(sys.argv) > 3) else 3)
        acceptors = self.new(Acceptor, num=nacceptors)
        proposers = self.new(Proposer, (acceptors,), num=nproposers)
        learners = self.new(Learner, (acceptors,), num=nlearners)
        for p in acceptors:
            self._setup(p, (learners,))
        self._start(((acceptors | proposers) | learners))
        super()._label('_st_label_779', block=False)
        l = None

        def UniversalOpExpr_780():
            nonlocal l
            for l in learners:
                if (not PatternExpr_795.match_iter(self._Node_ReceivedEvent_0, _BoundPattern801_=l)):
                    return False
            return True
        _st_label_779 = 0
        while (_st_label_779 == 0):
            _st_label_779 += 1
            if UniversalOpExpr_780():
                _st_label_779 += 1
            else:
                super()._label('_st_label_779', block=True)
                _st_label_779 -= 1
        self.output('done')
        self.send(('done',), to=(acceptors | proposers))



# -*- generated by 1.0.6 -*-
import da
PatternExpr_191 = da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern195_')])
PatternExpr_198 = da.pat.BoundPattern('_BoundPattern199_')
PatternExpr_223 = da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('abort'), da.pat.BoundPattern('_BoundPattern228_')])
PatternExpr_231 = da.pat.BoundPattern('_BoundPattern232_')
PatternExpr_253 = da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('ready'), da.pat.BoundPattern('_BoundPattern258_')])
PatternExpr_261 = da.pat.FreePattern('c')
PatternExpr_298 = da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('ready'), da.pat.BoundPattern('_BoundPattern303_')])
PatternExpr_306 = da.pat.BoundPattern('_BoundPattern307_')
PatternExpr_335 = da.pat.TuplePattern([da.pat.ConstantPattern('done'), da.pat.BoundPattern('_BoundPattern338_')])
PatternExpr_341 = da.pat.BoundPattern('_BoundPattern342_')
PatternExpr_233 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern239_')]), da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('abort'), da.pat.BoundPattern('_BoundPattern245_')])])
PatternExpr_308 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern314_')]), da.pat.TuplePattern([da.pat.ConstantPattern('vote'), da.pat.ConstantPattern('ready'), da.pat.BoundPattern('_BoundPattern320_')])])
PatternExpr_343 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.BoundPattern('_BoundPattern349_')]), da.pat.TuplePattern([da.pat.ConstantPattern('done'), da.pat.BoundPattern('_BoundPattern353_')])])
PatternExpr_405 = da.pat.TuplePattern([da.pat.ConstantPattern('prepare'), da.pat.FreePattern('tid')])
PatternExpr_412 = da.pat.FreePattern('coord')
PatternExpr_443 = da.pat.TuplePattern([da.pat.ConstantPattern('abort'), da.pat.FreePattern('tid')])
PatternExpr_456 = da.pat.TuplePattern([da.pat.ConstantPattern('commit'), da.pat.FreePattern('tid')])
PatternExpr_463 = da.pat.FreePattern('coord')
PatternExpr_518 = da.pat.TuplePattern([da.pat.ConstantPattern('done')])
PatternExpr_523 = da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.TuplePattern([da.pat.FreePattern(None), da.pat.FreePattern(None), da.pat.FreePattern(None)]), da.pat.TuplePattern([da.pat.ConstantPattern('done')])])
_config_object = {}
import sys
from random import randint

class Coordinator(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._CoordinatorReceivedEvent_0 = []
        self._CoordinatorReceivedEvent_1 = []
        self._CoordinatorReceivedEvent_2 = []
        self._CoordinatorReceivedEvent_3 = []
        self._CoordinatorReceivedEvent_4 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_0', PatternExpr_191, sources=[PatternExpr_198], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_1', PatternExpr_223, sources=[PatternExpr_231], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_2', PatternExpr_253, sources=[PatternExpr_261], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_3', PatternExpr_298, sources=[PatternExpr_306], destinations=None, timestamps=None, record_history=True, handlers=[]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CoordinatorReceivedEvent_4', PatternExpr_335, sources=[PatternExpr_341], destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, tid, cohorts, **rest_592):
        super().setup(tid=tid, cohorts=cohorts, **rest_592)
        self._state.tid = tid
        self._state.cohorts = cohorts
        pass

    def run(self):
        self.to_commit()
        self.send(('done',), to=self._state.cohorts)
        self.output('terminating')

    def to_commit(self):
        self.send(('prepare', self._state.tid), to=self._state.cohorts)
        super()._label('_st_label_182', block=False)
        c = None

        def UniversalOpExpr_183():
            nonlocal c
            for c in self._state.cohorts:

                def ExistentialOpExpr_189(c):
                    for (_, (_, _, _BoundPattern206_), (_ConstantPattern208_, _, _BoundPattern211_)) in self._CoordinatorReceivedEvent_0:
                        if (_BoundPattern206_ == c):
                            if (_ConstantPattern208_ == 'vote'):
                                if (_BoundPattern211_ == self._state.tid):
                                    if True:
                                        return True
                    return False
                if (not ExistentialOpExpr_189(c=c)):
                    return False
            return True
        _st_label_182 = 0
        while (_st_label_182 == 0):
            _st_label_182 += 1
            if UniversalOpExpr_183():
                _st_label_182 += 1
            else:
                super()._label('_st_label_182', block=True)
                _st_label_182 -= 1
        c = None

        def ExistentialOpExpr_216():
            nonlocal c
            for c in self._state.cohorts:
                if PatternExpr_233.match_iter(self._CoordinatorReceivedEvent_1, _BoundPattern239_=c, _BoundPattern245_=self._state.tid, SELF_ID=self._id):
                    return True
            return False
        if ExistentialOpExpr_216():
            s = {c for c in self._state.cohorts for (_, (_, _, _FreePattern269_), (_ConstantPattern271_, _ConstantPattern273_, _BoundPattern275_)) in self._CoordinatorReceivedEvent_2 if (_FreePattern269_ == c) if (_ConstantPattern271_ == 'vote') if (_ConstantPattern273_ == 'ready') if (_BoundPattern275_ == self._state.tid)}
            self.send(('abort', self._state.tid), to=s)
            self.abort(self._state.tid)
        c = None

        def UniversalOpExpr_291():
            nonlocal c
            for c in self._state.cohorts:
                if (not PatternExpr_308.match_iter(self._CoordinatorReceivedEvent_3, _BoundPattern314_=c, _BoundPattern320_=self._state.tid, SELF_ID=self._id)):
                    return False
            return True
        if UniversalOpExpr_291():
            self.send(('commit', self._state.tid), to=self._state.cohorts)
            super()._label('_st_label_327', block=False)
            c = None

            def UniversalOpExpr_328():
                nonlocal c
                for c in self._state.cohorts:
                    if (not PatternExpr_343.match_iter(self._CoordinatorReceivedEvent_4, _BoundPattern349_=c, _BoundPattern353_=self._state.tid, SELF_ID=self._id)):
                        return False
                return True
            _st_label_327 = 0
            while (_st_label_327 == 0):
                _st_label_327 += 1
                if UniversalOpExpr_328():
                    _st_label_327 += 1
                else:
                    super()._label('_st_label_327', block=True)
                    _st_label_327 -= 1
            self.commit(self._state.tid)

    def abort(self, tid):
        self.output(('abort:' + str(tid)))

    def commit(self, tid):
        self.output(('commit:' + str(tid)))

class Cohort(da.DistProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._CohortReceivedEvent_3 = []
        self._events.extend([da.pat.EventPattern(da.pat.ReceivedEvent, '_CohortReceivedEvent_0', PatternExpr_405, sources=[PatternExpr_412], destinations=None, timestamps=None, record_history=None, handlers=[self._Cohort_handler_404]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CohortReceivedEvent_1', PatternExpr_443, sources=None, destinations=None, timestamps=None, record_history=None, handlers=[self._Cohort_handler_442]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CohortReceivedEvent_2', PatternExpr_456, sources=[PatternExpr_463], destinations=None, timestamps=None, record_history=None, handlers=[self._Cohort_handler_455]), da.pat.EventPattern(da.pat.ReceivedEvent, '_CohortReceivedEvent_3', PatternExpr_518, sources=None, destinations=None, timestamps=None, record_history=True, handlers=[])])

    def setup(self, failure_rate, **rest_592):
        super().setup(failure_rate=failure_rate, **rest_592)
        self._state.failure_rate = failure_rate
        pass

    def run(self):
        super()._label('_st_label_515', block=False)
        _st_label_515 = 0
        while (_st_label_515 == 0):
            _st_label_515 += 1
            if PatternExpr_523.match_iter(self._CohortReceivedEvent_3, SELF_ID=self._id):
                _st_label_515 += 1
            else:
                super()._label('_st_label_515', block=True)
                _st_label_515 -= 1

    def prepared(self, tid):
        return (randint(0, 100) > self._state.failure_rate)

    def ready(self, tid):
        self.output(('ready:' + str(tid)))

    def abort(self, tid):
        self.output(('abort:' + str(tid)))

    def commit(self, tid):
        self.output(('commit:' + str(tid)))

    def _Cohort_handler_404(self, tid, coord):
        if self.prepared(tid):
            self.send(('vote', 'ready', tid), to=coord)
            self.ready(tid)
        else:
            self.send(('vote', 'abort', tid), to=coord)
            self.abort(tid)
    _Cohort_handler_404._labels = None
    _Cohort_handler_404._notlabels = None

    def _Cohort_handler_442(self, tid):
        self.abort(tid)
    _Cohort_handler_442._labels = None
    _Cohort_handler_442._notlabels = None

    def _Cohort_handler_455(self, tid, coord):
        self.send(('done', tid), to=coord)
        self.commit(tid)
    _Cohort_handler_455._labels = None
    _Cohort_handler_455._notlabels = None

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])

    def run(self):
        ncohorts = (int(sys.argv[1]) if (len(sys.argv) > 1) else 10)
        fail_rate = (int(sys.argv[2]) if (len(sys.argv) > 2) else 10)
        cohorts = self.new(Cohort, (fail_rate,), num=ncohorts)
        coordinators = self.new(Coordinator, ('001', cohorts), num=1)
        self._start((cohorts | coordinators))
